/* s4735.c

   Software for controlling the Si4735 based "PC radio
   with USB,  version 2 with 32 kHz xtal" from 
   AK MODL-BUS Computer GmbH 
   
   More info: http://www.elexs.de/SI4735i.html 

   This program is distributed under the GPL, version 2

   Copyright (C) 2011,2012 Tobias Stoeber
   Based on demo program / Delphi code (C) Burkhard Kainka
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; version 2.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#define INFO "20120507-0.1e Si4735 PC Radio with USB V2 Software (based on libftd2xx or libftdi) by Tobias Stoeber <tobi@to-st.de>\nLicence: GPLv2 - GNU GPL Version 2 <http://gnu.org/licenses/gpl.html>\nSi4735 PC Radio software comes with ABSOLUTELY NO WARRANTY\nThis is free software, and you are welcome to redistribute it under certain conditions.\n"
#define PROGRAM "Si4735 FM/AM Radio"
#define VERSION "0.1e-20120507"

#include <ctype.h>

#ifdef FTD2XX
#include <ftd2xx.h>
#else
#include <ftdi.h>
#define FT_STATUS	int
#define DWORD 		unsigned int
#define UCHAR		unsigned char
#define FT_OK		(0)
#define FT_ERROR	(-1)
#endif

#include <gtk/gtk.h>
#include <math.h>
#include <locale.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

/* gettext */
#include <libintl.h>
#define _(String) gettext (String)

/* RDS PTY definitions */
#include "rds_pty.h"

/* Basic RDS character set definitions */
#include "codetable.h"

/* some program defs */
#include "si4735.h"

/* small Si4735 IC bitmap 32x32 px */
#include "bmp-si4735-32x32.xpm"

/* pin defs */
#define PIN_TX (1 << 0)
#define PIN_RX (1 << 1)
#define PIN_RTS (1 << 2)
#define PIN_CTS (1 << 3)
#define PIN_DTR (1 << 4)
#define PIN_DCD (1 << 5)
#define PIN_CDR (1 << 6)
#define PIN_RI (1 << 7)

/* number of presets for AM and FM */
#define NUM_FM_PRESETS 48
#define NUM_AM_PRESETS 48

/* debugging */
/* #define DEBUG 0 */

#ifdef DEBUG
#define OUTPUT_DEBUG_STRING(str) printf(str);
#else
#define OUTPUT_DEBUG_STRING(str) ;
#endif

/* ============================
   .::. SI4735 RADIO PART .::. 
   ============================ */
  
unsigned char Codearray_RT_A[64];
unsigned char Codearray_RT_B[64]; 	
unsigned char CodearrayRT2[64];
unsigned char CodearrayPS[64];
unsigned char tmpCodearrayPS[64];
unsigned char CodearrayPTYN[64];
unsigned char FT_Out_Buffer[0x1000];
unsigned char FT_In_Buffer[0x1000];
int Fehler;
int FT_port = 7;
int Buf_adr = 0;
unsigned int R[14]; 
int DataOut[13] = { 0,0,0,0,0,0,0,0,0,0,0,0,0 };
int f_alt;
int f;
int c;
int Code;
double cap;
int rdsaktiv, rdsTimer, Zeit, UTCoffset, Minuten, Stunden;
int RDS_PI;
int TP, TA, MS, PTY, MS;
int autoTune;
int RxVolMuted = 0;
int RTflagAB = 0;
int RToldbit = 0;

int GTcnt[16] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };

int isModeFM = 0;
int isModeAM = 0;

/* debugging */
struct dbg_flags dbgFlag = { 0, 0, 0, 0, 0, 0}; /* debug Flags */

/* GPIO dev */
int isGPIOenabled = 0;
int GPOstate = 0;
int GPIOdelay = 35; 		/* delay between GPO pulses in ms */

/* misc start values, frequency, volume etc. */
int tunedFreq = 0;		/* currently tuned freq AM or FM */
int startFreqFM = 8750;		/* default freq  : 87.50 MHz */
int startFreqAM = 153;		/* default freq  : 153 kHz */
int startVolume = 58;		/* default volume: 58 */
int startRDS = 0;		/* auto start RDS decoding? */

int startRDSlog = 0;		/* stat RDS logfile */
char *fileRDSlog;		/* RDS logfile filename */
char *rdsLogFormat;		/* read RDS logfile format from stdin later */

struct rt_plus_tag RTplus1;
struct rt_plus_tag RTplus2;

int FM_Band_Selected = 0;
int FM_Band_Bottom[3] = { 8750, 7600, 6580 }; 
int FM_Band_Top[3] = { 10800, 9000, 7400 }; 

struct preset FM_Preset[ NUM_FM_PRESETS ];
int FM_Preset_Bank = 0;

struct preset AM_Preset[ NUM_AM_PRESETS ];
int AM_Preset_Bank = 0;

/* ============
   FTDI vars 
   ============ */

/* FTD2xx declarations */
#ifdef FTD2XX 
  FT_STATUS	ftStatus;
  FT_HANDLE	ftHandle;
  int ftPort = 0;
#else
  struct ftdi_context ftdic;
  int f; 
  int vid = 0x0403;
  int pid = 0x6001;
  int interface = INTERFACE_ANY;
  struct ftdi_device_list *devlist, *curdev;
  char manufacturer[128], description[128], serial[128];
#endif

  char *ser;
  int use_iSerial = 0;


/* =============
   misc unsorted 
   ============= */

unsigned char ucMode = 0x00;

/* global RDS text, vars */
char time_str[25] = "** RDS Zeitinfo **";

/* vars to hold misc info */
struct si_chip_info info_si;

/* ==========================
      Gtk vars 
   ========================== */
/* GTK declarations */
GtkWidget *window;
GtkWidget *notebook;
GtkWidget *hbox, *vbox, *bbox;
GtkWidget *main_vbox_fm, *main_vbox_rds, *main_vbox_am, *global_vbox;
GtkWidget *main_vbox_gp;
GtkWidget *hscrollbar_vol;
GtkWidget *button_quit, *button_exit;
GtkWidget *button_fm_start, *button_fm_seek_down, *button_fm_seek_up;
GtkWidget *button_am_start, *button_am_seek_down, *button_am_seek_up;
GtkWidget *button_am_tune_1up, *button_am_tune_5up, *button_am_tune_9up, *button_am_tune_10up;
GtkWidget *button_am_tune_1down, *button_am_tune_5down, *button_am_tune_9down, *button_am_tune_10down;
GtkWidget *button_gpio1_1, *button_gpio1_2;
GtkWidget *button_gpio1_1p, *button_gpio1_2p, *button_gpio1_3p, *button_gpio1_4p;
GtkWidget *frame_fm_tune, *frame_vol, *frame_fm_rds, *frame_fm_smeter;
GtkWidget *frame_am_tune, *frame_am_preset, *frame_am_band, *frame_am_smeter;
GtkWidget *frame_gpio1, *frame_gpio2, *frame_gpio3;
GtkWidget *label;
/* GTK: notebook */
GtkWidget *notebook_label_fm;
GtkWidget *notebook_label_rds;
GtkWidget *notebook_label_am;
GtkWidget *notebook_label_gp;
/* GTK: statusbar */
GtkWidget *statusbar;
/* GTK: toolbar */
GtkWidget *toolbar;
GtkWidget *tb_vbox;
// GtkToolItem *tb_open;
// GtkToolItem *tb_save;
GtkToolItem *tb_screenshot;
GtkToolItem *tb_info;
GtkToolItem *tb_sep;
GtkToolItem *tb_prefs;
GtkToolItem *tb_about;
GtkToolItem *tb_help;
GtkToolItem *tb_exit;

GtkWidget *button_mute;
GtkWidget *button_fm_smeter_update, *button_am_smeter_update;   
GtkWidget *button_rds;
GtkWidget *entry_fm_freq, *entry_am_freq;
GtkWidget *entry_rds_psn, *entry_rds_pi, *entry_rds_rt_a, *entry_rds_rt_b, *entry_rds_af, *entry_rds_time;
GtkWidget *label_rds_psn, *label_rds_pi, *label_rds_rt_a, *label_rds_rt_b, *label_rds_af, *label_rds_time;
GtkWidget *entry_rds_pty, *label_rds_pty, *entry_rds_ms, *entry_rds_ta, *entry_rds_tp;
GtkWidget *entry_fm_smeter_rssi, *entry_fm_smeter_snr, *entry_am_smeter_rssi, *entry_am_smeter_snr, *entry_am_smeter_cap, *entry_am_smeter_ind; 
GtkWidget *label_fm_smeter_rssi, *label_fm_smeter_snr, *label_am_smeter_rssi, *label_am_smeter_snr, *label_am_smeter_cap, *label_am_smeter_ind;
GtkWidget *button_fm_band_norm, *button_fm_band_jp, *button_fm_band_oirt;
GtkWidget *button_fm_preset[6], *button_fm_preset_bank;
GtkWidget *button_am_preset[6], *button_am_preset_bank;
GtkWidget *button_am_band[15];
GtkWidget *button_fm_autoscan, *button_am_autoscan; 
GtkObject *adj_fm_tune, *adj_am_tune, *adj_vol;

GtkTooltips *button_tips;
GtkTooltips *entry_tips;
GtkTooltips *prefs_tips;

/* colors for text entry backgrounds M/S, TP, TA */
GdkColor colorWhite = { 0, 65535, 65535, 65535 };
GdkColor colorGreen = { 1, 0, 65535, 0 };

/* GTK: misc */
// gboolean _isSensitive = FALSE;

/* declare stubs */
void AM_GPIO_Set(int freq);

/* ===========================
       some helper functs
   =========================== */
   
/* get hi part */ 
int hi(int d)
{
  int ret = d >> 8;
  return ret;
}

/* get low part */
int lo(int d)
{
  int ret = ( d & 0b0000000011111111 );
  return ret;
}

/* delay in ms */
void Delay(int t)
{
  usleep(t * 1000);
}

char* substring(const char* str, size_t begin, size_t len)
{
  if (str == 0 || strlen(str) == 0 || strlen(str) < begin || strlen(str) < (begin+len))
    return 0;

  return strndup(str + begin, len);
}

/* ================================================
   read/save application settings and presets
   ================================================ */

gboolean si4735radio_settings_load (void)
{
  GKeyFile *key_file;
  gboolean res;
  gchar *path;

  int i;
  int am_arr_size = sizeof( SI4735_prop_am )/sizeof(  SI4735_prop_am[0]);
  int fm_arr_size = sizeof( SI4735_prop_fm )/sizeof(  SI4735_prop_fm[0]);

  path = g_build_path (G_DIR_SEPARATOR_S,
	               g_get_user_config_dir(),
	               "si4735radio",
                       "application-settings",
                       NULL);

  key_file = g_key_file_new();
  res = g_key_file_load_from_file (key_file,
                                   path,
                                   G_KEY_FILE_KEEP_COMMENTS|G_KEY_FILE_KEEP_TRANSLATIONS,
                                   NULL);
  g_free (path);
  if (!res)
    return res;

  /* AM properties */
  for ( i = 0; i < am_arr_size; i++ )
  {
    if ( (SI4735_prop_am[i].val_curr == -1) && (SI4735_prop_am[i].prop > 0) )
    {
      SI4735_prop_am[i].val_curr  = g_key_file_get_integer (key_file,
                                                            "AM Properties",
                                                            SI4735_prop_am[i].name,
                                                            NULL);
    }
  }

  /* FM properties */
  for ( i = 0; i < fm_arr_size; i++ )
  {
    if ( (SI4735_prop_fm[i].val_curr == -1) && (SI4735_prop_fm[i].prop > 0) )
    {
      SI4735_prop_fm[i].val_curr  = g_key_file_get_integer (key_file,
                                                            "FM Properties",
                                                            SI4735_prop_fm[i].name,
                                                            NULL);
    }
  }
  
  g_key_file_free (key_file);

  return TRUE;
}

gboolean si4735radio_settings_save (void)
{
  gchar *contents = NULL;
  gsize content_size = 0;
  GKeyFile *key_file;
  gboolean res;
  gchar *path;

  int i;
  int am_arr_size = sizeof( SI4735_prop_am )/sizeof(  SI4735_prop_am[0]);
  int fm_arr_size = sizeof( SI4735_prop_fm )/sizeof(  SI4735_prop_fm[0]);

  /* create config dir pathname */
  path = g_build_path (G_DIR_SEPARATOR_S,
                        g_get_user_config_dir (),
                        "si4735radio",
                        NULL);

  /* check if apps config dir exists, create if necessary */
  struct stat d_stat;
  
  if ( stat(path, &d_stat) != 0)
  {
    /* directory does not exist */
    if ( mkdir(path, 00755) != 0)
    {
      fprintf(stderr,"error creating dir: %s\n", path);
    }
  }                                  

  /* create presets path + filename */
  path = g_build_path (G_DIR_SEPARATOR_S,
                        g_get_user_config_dir (),
                        "si4735radio",
                        "application-settings",
                        NULL);

  key_file = g_key_file_new ();
  res = g_key_file_load_from_file (key_file,
                                    path,
                                    G_KEY_FILE_KEEP_COMMENTS|G_KEY_FILE_KEEP_TRANSLATIONS,
                                    NULL);

  /* AM properties */
  for ( i = 0; i < am_arr_size; i++ )
  {
    if ( (SI4735_prop_am[i].val_curr > -1) && (SI4735_prop_am[i].prop > 0) )
    {
      g_key_file_set_integer (key_file,
                              "AM Properties",
                              SI4735_prop_am[i].name,
                              SI4735_prop_am[i].val_curr);
    }
  }

  /* FM properties */
  for ( i = 0; i < fm_arr_size; i++ )
  {
    if ( (SI4735_prop_fm[i].val_curr > -1) && (SI4735_prop_fm[i].prop > 0) )
    {
      g_key_file_set_integer (key_file,
                              "FM Properties",
                              SI4735_prop_fm[i].name,
                              SI4735_prop_fm[i].val_curr);
    }
  }

  contents = g_key_file_to_data (key_file, &content_size, NULL);
  g_file_set_contents (path, contents, content_size, NULL);
  g_free (contents);
  g_free (path);

  g_key_file_free (key_file);

  return TRUE;
}

gboolean si4735radio_presets_load (void)
{
  int i;
  
  GKeyFile *key_file;
  gboolean res;
  gchar *path;
  gchar *station_name;

  path = g_build_path (G_DIR_SEPARATOR_S,
	               g_get_user_config_dir(),
	               "si4735radio",
                       "presets",
                       NULL);

  key_file = g_key_file_new();
  res = g_key_file_load_from_file (key_file,
                                   path,
                                   G_KEY_FILE_KEEP_COMMENTS|G_KEY_FILE_KEEP_TRANSLATIONS,
                                   NULL);
  g_free (path);
  if (!res)
    return res;

  /* -------------
     FM presets 
     ------------- */
  for ( i = 0; i < NUM_FM_PRESETS; i++ )
  {
    gchar group_name[32];
    sprintf(group_name, "FM Preset %02d", i);

    if (  g_key_file_has_group( key_file, group_name ) )
    {
      FM_Preset[i].freq = g_key_file_get_integer(key_file,
                                                  group_name,
                                                  "freq",
                                                  NULL);

      station_name = g_key_file_get_string(key_file,
                                            group_name,
                                            "station-name",
                                            NULL);

      strcpy(FM_Preset[i].station_name, (char *)station_name);
    
      FM_Preset[i].rds_pi = g_key_file_get_integer(key_file,
                                                    group_name,
                                                    "rds-pi",
                                                    NULL);
    }
  }

  /* -------------
     AM presets 
     ------------- */
  for ( i = 0; i < NUM_AM_PRESETS; i++ )
  {
    gchar group_name[32];
    sprintf(group_name, "AM Preset %02d", i);

    if (  g_key_file_has_group( key_file, group_name ) )
    {
      AM_Preset[i].freq = g_key_file_get_integer(key_file,
                                                  group_name,
                                                  "freq",
                                                  NULL);

      station_name = g_key_file_get_string(key_file,
                                            group_name,
                                            "station-name",
                                            NULL);

      strcpy(AM_Preset[i].station_name, (char *)station_name);
    }
  }

  g_key_file_free (key_file);

  return TRUE;
}

gboolean si4735radio_presets_save (void)
{
  int i;
  
  gchar *contents = NULL;
  gsize content_size = 0;
  GKeyFile *key_file;
  gboolean res;
  gchar *path;

  /* create config dir pathname */
  path = g_build_path (G_DIR_SEPARATOR_S,
                        g_get_user_config_dir (),
                        "si4735radio",
                        NULL);

  /* check if apps config dir exists, create if necessary */
  struct stat d_stat;
  
  if ( stat(path, &d_stat) != 0)
  {
    /* directory does not exist */
    if ( mkdir(path, 00755) != 0)
    {
      fprintf(stderr,"error creating dir: %s\n", path);
    }
  }                                  

  /* create presets path + filename */
  path = g_build_path (G_DIR_SEPARATOR_S,
                        g_get_user_config_dir (),
                        "si4735radio",
                        "presets",
                        NULL);

  key_file = g_key_file_new ();
  res = g_key_file_load_from_file (key_file,
                                    path,
                                    G_KEY_FILE_KEEP_COMMENTS|G_KEY_FILE_KEEP_TRANSLATIONS,
                                    NULL);

  /* -------------
     FM presets 
     ------------- */
  for ( i = 0; i < NUM_FM_PRESETS; i++ )
  {
    gchar group_name[32];
    sprintf(group_name, "FM Preset %02d", i);

    g_key_file_set_integer(key_file,
                            group_name,
                            "freq",
                            FM_Preset[i].freq);

    g_key_file_set_string(key_file,
                          group_name,
                          "station-name",
                          (char *) FM_Preset[i].station_name);

    g_key_file_set_integer(key_file,
                            group_name,
                            "rds-pi",
                            FM_Preset[i].rds_pi);
  }

  /* -------------
     AM presets 
     ------------- */
  for ( i = 0; i < NUM_AM_PRESETS; i++ )
  {
    gchar group_name[32];
    sprintf(group_name, "AM Preset %02d", i);

    g_key_file_set_integer(key_file,
                            group_name,
                            "freq",
                            AM_Preset[i].freq);

    g_key_file_set_string(key_file,
                          group_name,
                          "station-name",
                          (char *) AM_Preset[i].station_name);

    g_key_file_set_integer(key_file,
                            group_name,
                            "rds-pi",
                            AM_Preset[i].rds_pi);
  }

  contents = g_key_file_to_data (key_file, &content_size, NULL);
  g_file_set_contents (path, contents, content_size, NULL);
  g_free (contents);
  g_free (path);

  g_key_file_free (key_file);

  return TRUE;
}

/* ================================================
   FTDI routines
   ================================================ */

int FTDI_Init(void)
{
  /* Initialize, open (first) device */
#ifdef FTD2XX
  if (FT_Open(ftPort, &ftHandle) != FT_OK)
  {
    printf("Can't open device.\nFTOpen(%d) failed.\n", ftPort);
    puts("Either no device connected or wrong permissions.\nThis may also fail, if ftdi_sio driver is loaded.\nPlease refer to file USAGE for troubleshooting tips.");
    return 1;
  }
#else
  if (ftdi_init(&ftdic) < 0)
  {
    fprintf(stderr, "ftdi_init failed\n");
    return EXIT_FAILURE;
  }

  /* select first interface */
  ftdi_set_interface(&ftdic, interface);

  /* open the device */
  if ( use_iSerial == 1 )
  {
     char devstr[32];
     sprintf(devstr, "s:0x%04x:0x%04x:%s", vid, pid, ser);
     
     f = ftdi_usb_open_string(&ftdic, devstr);
  }
  else
  {
    f = ftdi_usb_open(&ftdic, vid, pid);
  }
  
  if (f < 0)
  {
    fprintf(stderr, "unable to open ftdi device: %d (%s)\n", f, ftdi_get_error_string(&ftdic));
    exit(-1);
  }
#endif
  return EXIT_SUCCESS;
}

int FTDI_Connect(void)
{ 
  /* set sync bitbang mode */
#ifdef FTD2XX
  ftStatus = FT_SetBitMode(ftHandle, PIN_RTS | PIN_DTR, 4);
  if (ftStatus != FT_OK)
  {
    printf("Failed to set bit mode.\n");
    return 1;
  }
#else
    if (ftdi_set_bitmode(&ftdic, PIN_RTS | PIN_DTR, BITMODE_SYNCBB) < 0)
            fprintf(stderr, "Can't enable bitbang");
#endif

  /* set BaudRate */
int baudrate = 9600;
#ifdef FTD2XX
  ftStatus = FT_SetBaudRate(ftHandle, baudrate);  /* Actually was 9600 * 16 */
  if (ftStatus != FT_OK)
  {
    printf("Failed to set baud rate.\n");
    return 1;
  }
#else
  // Set baudrate
  f = ftdi_set_baudrate(&ftdic, baudrate);
  if (f < 0)
  {
    fprintf(stderr, "unable to set baudrate: %d (%s)\n", f, ftdi_get_error_string(&ftdic));
    exit(-1);
  }
#endif

  /* set SetUSBParameters */
  /* USB request transfer size, 4k default, must be between 64 byte und 64 kByte */
#ifdef FTD2XX
  ftStatus = FT_SetUSBParameters(ftHandle, 4096, 4096);
  if (ftStatus != FT_OK)
  {
    printf("Failed to set usb parameters.\n");
    return 1;
  }
#else
  ftdi_read_data_set_chunksize(&ftdic, 4096);
  ftdi_set_latency_timer(&ftdic, 2); // This helps make reads much faster.  16 appeared to be the default.
#endif


  return EXIT_SUCCESS;  
}

int FTDI_Find_All(int on_exit)
{
  int ret, i;

#ifdef FTD2XX
  if (FT_Open(ftPort, &ftHandle) != FT_OK)
  {
    printf("Can't open device.\nFTOpen(%d) failed.\n", ftPort);
    puts("Either no device connected or wrong permissions.\nThis may also fail, if ftdi_sio driver is loaded.\nPlease refer to file USAGE for troubleshooting tips.");
    return 1;
  }
#else
  if (ftdi_init(&ftdic) < 0)
  {
    fprintf(stderr, "ftdi_init failed\n");
    return EXIT_FAILURE;
  }
#endif

#ifdef FTD2XX
  // do nothing at the moment
#else
  if ((ret = ftdi_usb_find_all(&ftdic, &devlist, vid, pid)) < 0)
  {
      fprintf(stderr, "ftdi_usb_find_all failed: %d (%s)\n", ret, ftdi_get_error_string(&ftdic));
      return EXIT_FAILURE;
  }

  printf("Number of FTDI devices found: %d\n", ret);

  i = 0;

  for (curdev = devlist; curdev != NULL; i++)
  {
    printf("Checking device: %d\n", i);
    if ((ret = ftdi_usb_get_strings(&ftdic, curdev->dev, manufacturer, 128, description, 128, serial, 128)) < 0)
    {
      fprintf(stderr, "ftdi_usb_get_strings failed: %d (%s)\n", ret, ftdi_get_error_string(&ftdic));
      return EXIT_FAILURE;
    }
    printf("Manufacturer: %s, Description: %s, Serial: %s\n", manufacturer, description, serial);

    // Read out FTDIChip-ID of R type chips
// #ifdef FTD2XX
    // do nothing
    // FT_Close(ftHandle);
// #else
    if (ftdi_usb_open(&ftdic, vid, pid) < 0)
    {
      fprintf(stderr, "unable to open ftdi device: (%s)\n", ftdi_get_error_string(&ftdic));
    }

    if (ftdic.type == TYPE_R)
    {
      unsigned int chipid;
      printf("ftdi_read_chipid: %d\n", ftdi_read_chipid(&ftdic, &chipid));
      printf("FTDI chipid: %X\n", chipid);
    }

    ftdi_usb_close(&ftdic);
// #endif

    printf("\n");
    curdev = curdev->next;
  }

  ftdi_list_free(&devlist);
  ftdi_deinit(&ftdic);
#endif

  if ( on_exit == -1 )
    exit(0);

  return EXIT_SUCCESS;
}

int Write_USB_Device_Buffer(DWORD Write_Count)
{
  /* Writes Write_Count Bytes from FT_Out_Buffer to the USB device
   * Function returns the number of bytes actually sent
   * In this example, Write_Count should be 32k bytes max
   */
  DWORD Write_Result = 0;

#ifdef FTD2XX   
  Write_Result = FT_Write(ftHandle, FT_Out_Buffer, Write_Count, &Write_Result);
#else
  Write_Result = ftdi_write_data(&ftdic, FT_Out_Buffer, Write_Count);
  
  if ( Write_Result < 0)
  {
    fprintf(stderr,"write failed for 0x%x: %s\n",
                    f, ftdi_get_error_string(&ftdic));
  }
#endif

  return (Write_Result);
}

int Read_USB_Device_Buffer(DWORD Read_Count) 
{
  /* Reads Read_Count Bytes ( or less ) from the USB device to the FT_In_Buffer
   * Function returns the number of bytes actually received  which may range from zero
   * to the actual number of bytes requested, depending on how many have been received
   * at the time of the request + the read timeout value.
   */

  DWORD Read_Result = 0;

#ifdef FTD2XX
  Read_Result = FT_Read(ftHandle, FT_In_Buffer, Read_Count, &Read_Result);
#else
  Read_Result = ftdi_read_data(&ftdic, FT_In_Buffer, sizeof(FT_In_Buffer)); // Read_Count);
#endif
  return(Read_Result);
}

/* ================================================
   I2C Routines
   ================================================ */

void SCL(int d)
{
  if ( d == 0 ) {
    FT_port = FT_port & (255-4);
  } else {
    FT_port = FT_port | (4); /* Bit 2 RTS */
  }

  FT_Out_Buffer[Buf_adr] = FT_port;
  Buf_adr = Buf_adr + 1;
}

void SDA(int d)
{
  if ( d == 0 ) {
    FT_port = FT_port & (255-16);
  } else {
    FT_port = FT_port | (16); /* Bit 4 */
  }

  FT_Out_Buffer[Buf_adr] = FT_port;
  Buf_adr = Buf_adr + 1;
}

void I2C_Init()
{
  Buf_adr = 0;
  SCL(1); SDA(1);          /* SCL=1, SDA=1 */
}

void Start()
{
  SDA(0);          /* SDA=0 */
  SCL(0);          /* SCL=0 */
}

void Stop()
{
  SCL(0); SDA(0);  /* SCL=0, SDA=0 */
  SCL(1);          /* SCL=1 */
  SDA(1);          /* SDA=1 */ 
}

void Acknowledge()
{
  SCL(0); SDA(0);  /* SCL=0, SDA=0 */ 
  SCL(1);          /* SCL=1 */
  SCL(0);          /* SCL=0 */ 
}

void KeinAcknowledge()
{
  SCL(0); SDA(1);  /* SCL=0, SDA=1 */
  SCL(1);          /* SCL=1 */
  SCL(0);          /* SCL=0 */
}

void ChipReset()
{
  OUTPUT_DEBUG_STRING("ChipReset();\n");

  FT_Out_Buffer[0] = 20;
  Buf_adr = 1;
  Write_USB_Device_Buffer( Buf_adr);
  Read_USB_Device_Buffer( Buf_adr);
  Delay(10);
  FT_Out_Buffer[0] = 21;
  Buf_adr = 1;
  Write_USB_Device_Buffer( Buf_adr);
  Read_USB_Device_Buffer( Buf_adr);
  Delay(10);
}

void Ausgeben(int Wert)
{
  OUTPUT_DEBUG_STRING( "Ausgeben()\n" );

  int Bitwert, n;

  Bitwert = 128;

  for ( n = 1; n <= 8; n++ )
  {
    if ( Bitwert == (Wert & Bitwert) )
    { 
      SDA(1);
    }
    else
    {
      SDA(0);   /* SDA setzen */ 
    }

    SCL(1);     /* SCL=1 */
    SCL(0);     /* SCL=0 */

    Bitwert = Bitwert / 2;
  }

  SDA(1);           /* SDA=1 */
  SCL(1);           /* SCL=1, SDA lesen */
  SCL(0);           /* SCL=0 */ 
}

int Readi2c(int Ack)
{
  OUTPUT_DEBUG_STRING("Readi2c()\n");

  int n;
  int d;

  Buf_adr = 0;
  SDA(1);
  
  for ( n = 1; n <= 8; n++ )
  {
     SCL(1);
     SCL(0);
  }

  SDA(Ack);
  SCL(1);
  SCL(0);
  
  Write_USB_Device_Buffer( Buf_adr);
  Read_USB_Device_Buffer( Buf_adr);

  d = 0;

  for ( n = 1; n <= 8; n++ )
  {
    d = d * 2;

    if ( (FT_In_Buffer[0+2*n] & 8) > 0 )
    {
      d = d + 1;
    }
  }

  return d;
}

void I2Cio(int Nrout, int Nrin)
{
  OUTPUT_DEBUG_STRING( "Begin I2Cio()\n" );

  int n;

  Buf_adr = 0;
  I2C_Init();
  Start();

  Ausgeben(34);

  for ( n = 1; n <= Nrout; n++)
  {
    Ausgeben( DataOut[n] );
  }

  Stop();

  Write_USB_Device_Buffer( Buf_adr);

  Read_USB_Device_Buffer( Buf_adr);

  if ( Nrin > 0 )
  {
    Buf_adr = 0;
    Start();
    
    Ausgeben(35);

    Write_USB_Device_Buffer( Buf_adr);
    Read_USB_Device_Buffer( Buf_adr);

    for ( n = 0; n <= (Nrin-1); n++ )
    {
      R[n] = Readi2c(0);
    }

    R[Nrin] = Readi2c(1);

    Buf_adr = 0;
    Stop();

    Write_USB_Device_Buffer( Buf_adr);
    Read_USB_Device_Buffer( Buf_adr);
  }
}

/* ================================================
   Si4735 PC Radio USB Routines
   ================================================ */

int GET_INT_STATUS()
{
  int status;
  
  DataOut[1] = 0x14;
  I2Cio (1,1);
  status = Readi2c(1);   /* Nack */

  if ( (status & 128) == 128 )
  { 
    return 1;
  }
  else
  {
    return 0;
  }
}

int GET_RDS_STATUS()
{
  int status;
  
  DataOut[1] = 0x14;
  I2Cio (1,1);
  status = Readi2c(1);   /* Nack */

  if ( (status & 4) == 4 )
  { 
    return 1;
  }
  else
  {
    return 0;
  }
}

void waitCTS()
{
  int timeout;
  
  if ( ! GET_INT_STATUS() )
  {
     timeout = 0;
     do {
        timeout = timeout + 1;
     } while ( (! GET_INT_STATUS()) && (! (timeout > 1)));
  }
}

void Prop(int Adr, int Data)
{
  OUTPUT_DEBUG_STRING( "Prop()\n" );

  waitCTS();
  DataOut[1] = 0x12;
  DataOut[2] = 0x00;
  DataOut[3] = hi(Adr);
  DataOut[4] = lo(Adr);
  DataOut[5] = hi(Data);
  DataOut[6] = lo(Data);
  I2Cio(6,1);

  // printf("Prop 0x%x to 0x%x (%d)= 0%x\n", Adr, Data, Data, R[0]);
}

int Get_Prop(int Adr)
{
  OUTPUT_DEBUG_STRING( "Get_Prop()\n" );

  waitCTS();
  DataOut[1] = 0x13;
  DataOut[2] = 0x00;
  DataOut[3] = hi(Adr);
  DataOut[4] = lo(Adr);
  I2Cio(4,4);

  // printf("Get Prop 0x%x = %x%x  (%d)\n", Adr, R[2], R[3], R[2] * 256 + R[3]);
  return(R[2] * 256 + R[3]);
}

void Init_RDS()
{
  OUTPUT_DEBUG_STRING( "Init_RDS()\n" );

  Prop(0x1502, 0xAA01); /* {0,0,0,0} = 0x0001, {2,2,2,2} = 0xAA01, {3,2,3,3} = 0xEF01, {3,3,3,3} = 0xFF01 */

  Delay(50);
  
  Prop(0x1500, 0x0000);

  Delay(50);

  Prop(0x1501, 0x0000);
}

/* function Xtal(); actually not needed anymore */
void Xtal()
{
  OUTPUT_DEBUG_STRING( "Xtal()\n" );

  Prop(0x0202, 125);   /* 4 MHz / 125 = 32 kHz */
  Delay(10);
  Prop(0x0201, 32000);
}

void RX_VOLUME (int L)
{
  OUTPUT_DEBUG_STRING( "RX_Volume()\n" );

  Prop(0x4000, L);
}

void PowerDown()
{
  OUTPUT_DEBUG_STRING( "PowerDown()\n" );

  waitCTS();
  DataOut[1] = 0x11;
  I2Cio(1,0);
}

void GetRev()
{
  OUTPUT_DEBUG_STRING( "GetRev()\n" );

  waitCTS();
  DataOut[1] = 0x10;
  I2Cio(1,8);
}

void Init_AM()
{
  waitCTS();
  DataOut[1] = 0x01;
  DataOut[2] = 0x11;  /* 32 kHz */
  DataOut[3] = 0x05;
  I2Cio(3,0);
  Delay(550);
}

void Init_FM()
{
  waitCTS();
  DataOut[1] = 0x01;
  DataOut[2] = 0x10;   /* 32 kHz */
  DataOut[3] = 0x05;
  I2Cio(3,0);
  Delay(550);
}

void AM_Tune_Freq(int f)
{
  OUTPUT_DEBUG_STRING("AM_Tune_Freq()\n");

  waitCTS();
  DataOut[1] = 0x40;
  DataOut[2] = 0x00;
  DataOut[3] = hi(f);
  DataOut[4] = lo(f);
  DataOut[5] = 0x00;
  DataOut[6] = 0x00;
  I2Cio(6,0);
}

void FM_Tune_Freq(int f)
{
  OUTPUT_DEBUG_STRING("FM_Tune_Freq()\n");

  waitCTS();
  DataOut[1] = 0x20;
  DataOut[2] = 0x00;
  DataOut[3] = hi(f);
  DataOut[4] = lo(f);
  DataOut[5] = 0x00;
  DataOut[6] = 0x00;
  I2Cio(6,0);
}

void AM_Seek_Freq(int Arg1)
{
  OUTPUT_DEBUG_STRING("AM_Seek_Freq()\n");

  waitCTS();
  DataOut[1] = 0x41;
  DataOut[2] = Arg1;
  I2Cio(2,0);
}

void FM_Seek_Freq(int Arg1)
{
  OUTPUT_DEBUG_STRING("FM_Seek_Freq()\n");

  waitCTS();
  DataOut[1] = 0x21;
  DataOut[2] = Arg1;
  I2Cio(2,0);
}

void AM_Tune_Status(int Arg1)
{
  OUTPUT_DEBUG_STRING("AM_Tune_Status()\n");

  waitCTS();
  DataOut[1] = 0x42;
  DataOut[2] = Arg1;
  I2Cio(2,8);

  int f_tuned = (R[2] * 256 + R[3]);
  
  if ( f_tuned > 0 )
  {
    char lt[11];
    sprintf(lt, "%d", f_tuned);

    gtk_entry_set_text(GTK_ENTRY(entry_am_freq), lt);


    startFreqAM = f_tuned;
    tunedFreq = f_tuned;

    isModeFM = 0;
    isModeAM = 1;
  }
}

void FM_RDS_STATUS()
{
  OUTPUT_DEBUG_STRING("FM_RDS_Status()\n");

  waitCTS();
  DataOut[1] = 0x24;
  DataOut[2] = 0x01;
  I2Cio(2,13);
}

void AM_RSQ_Status()
{
  OUTPUT_DEBUG_STRING("AM_RSQ_Status()\n");

  waitCTS();
  DataOut[1] = 0x43;
  DataOut[2] = 0x00;
  I2Cio(2,5);
}

void FM_Tune_Status(int Arg1)
{
  OUTPUT_DEBUG_STRING("FM_Tune_Status()\n");

  waitCTS();
  DataOut[1] = 0x22;
  DataOut[2] = Arg1;
  I2Cio(2,8);

  int f_tuned = (R[2] * 256 + R[3]);

  if ( f_tuned > 0 )
  {
    char lt[11];
    sprintf(lt, "%.2f", (double) f_tuned / 100);

    gtk_entry_set_text(GTK_ENTRY(entry_fm_freq), lt);

    startFreqFM = f_tuned;
    tunedFreq = f_tuned;
    
    isModeFM = 1;
    isModeAM = 0;
  }
}

void FM_RSQ_Status()
{
  OUTPUT_DEBUG_STRING( "FM_RSQ_Status()\n" );

  waitCTS();
  DataOut[1] = 0x23;
  DataOut[2] = 0x00;
  I2Cio(2,5);
}


void FM_SMeter()
{
  if ( GET_INT_STATUS() )
  {
    FM_RSQ_Status();

    char rssi_str[12];
    sprintf(rssi_str, "%d dBµV", R[4]);

    char snr_str[12];
    sprintf(snr_str, "%d dB", R[5]);

    gtk_entry_set_text(GTK_ENTRY(entry_fm_smeter_rssi), (char *)rssi_str);
    gtk_entry_set_text(GTK_ENTRY(entry_fm_smeter_snr), (char *)snr_str);
  }
}

void AM_SMeter()
{
  int _rssi = 0, _snr = 0, _cap = 0, _f = 0;
  double _fcap = 0.0;
  double _ind = 0.0;
  
  if ( GET_INT_STATUS() )
  {
    AM_RSQ_Status();

    _rssi = R[4];
    _snr = R[5];
  }

  if ( GET_INT_STATUS() )
  {
    waitCTS();
    AM_Tune_Status(3);

    _f = 256 * R[2] + R[3];
    _cap = 256 * R[6] + R[7];

    _fcap = _cap * 0.095 + 7;
    _cap = (int)round (_fcap);

    if ( (_cap > 10) && (_cap < 591) )
    {
      _ind = (1 / ( 4 * M_PI * M_PI * pow(( _f * 1000),2) * ( _cap * pow(10,-12) ) )) * pow(10,6); 
    }
/*
    int arr_len = sizeof(AM_bands)/sizeof(AM_bands[0]);
    int p;

    for ( p = 0; p < arr_len; p++ )
    {
      if ( (_f >= AM_bands[p].band_bottom) && (_f <= AM_bands[p].band_top) )
      {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_am_band[p]), TRUE);
      }
      else
      {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_am_band[p]), FALSE);
      }
    }
*/
  }

  char rssi_str[12];
  sprintf(rssi_str, "%d dBµV", _rssi);

  char snr_str[12];
  sprintf(snr_str, "%d dB", _snr);

  char cap_str[12];
  sprintf(cap_str, "%d pF", _cap);

  char ind_str[12];
  if ( _ind > 0 )
  {
    sprintf(ind_str, "%.2f µH", _ind);
  }
  else
  {
    sprintf(ind_str, "n/a µH");
  }
  
  gtk_entry_set_text(GTK_ENTRY(entry_am_smeter_rssi), (char *)rssi_str);
  gtk_entry_set_text(GTK_ENTRY(entry_am_smeter_snr), (char *)snr_str);
  gtk_entry_set_text(GTK_ENTRY(entry_am_smeter_cap), (char *)cap_str);
  gtk_entry_set_text(GTK_ENTRY(entry_am_smeter_ind), (char *)ind_str);
}

/* ================================================
   RDS Routines
   ================================================ */

unsigned char * rds_charset_decode(unsigned char *a)
{
  int i,c;
  int len = 64;
  
  unsigned char *tmp = malloc(128);

  for ( i = 0; i < 128; i++ )
  {
    tmp[i] = 0x20;
  }

  tmp[0] = '\0';
  
  for ( i = 0; i < len; i++ )
  {
    c = (int)a[i];
    c = c - 0x20;
    
    if ( c < sizeof(codetable_rds) )
    {
      strcat((char *)tmp, (char *)codetable_rds[c]);
    }
  }
  
  return(tmp);  
}

void RDSupdate()
{
  FILE *fp;
  int n, pos, BLED;  int i; // max;

  /* vars for group 1A */
  int altFreq, altFreq2, numFreq, AM, noAFs;
  altFreq = 0; altFreq2 = 0;
  numFreq = 0;
  noAFs = 0;
  AM = 0;

  /* vars for group 4A */
  Minuten = 0;
  Stunden = 0;
  Zeit = 0;

  /* vars for TMC, group 8A */
  int TMC_T = -1;
  int TMC_F = -1;
  int TMC_DP = -1;
  int TMC_D = -1;
  int TMC_R = -1;
  int TMC_Extent = -1;
  int TMC_Event = -1;
  int TMC_Loc = -1;
  
  /* vars for group 1A */
  int variant;
  
  for ( n = 1; n <= 4; n++ )
  {
    /* Application.ProcessMessages replacement in Gtk */
    while ( gtk_events_pending() )
    {
      gtk_main_iteration();
    }

//    if ( GET_INT_STATUS() )
//    {
      FM_RDS_STATUS();
//    }

    /* open rds logfile for output */
    if ( startRDSlog == 1 )
    {
      if ( (fp = fopen(fileRDSlog, "a")) == NULL )
      {
        printf("Cannot open RDS log file.\n");
      }
      else
      {
        if ( strcmp("raw", rdsLogFormat) == 0 )
        {
          fprintf(fp, "%c%c%c%c %c%c %c%c %c%c %c%c %c%c\n", R[0],R[1],R[2],R[3], R[4],R[5],R[6],R[7], R[8],R[9],R[10],R[11], R[12],R[13]);
        }
        else
        {
          fprintf(fp, "%02X%02X%02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X\n", R[0],R[1],R[2],R[3], R[4],R[5],R[6],R[7], R[8],R[9],R[10],R[11], R[12],R[13]);
        }
        fclose(fp);
      }
    }
    
    // printf("RDS - Group Lost?: %s\n", (char *)(((R[2] & 4)/4) == 1 ? "yes\0" : "no\0") );
    // printf("RDS - Num of blocks remaining: %d\n", R[3]);

    if ( R[3] > 0 )
    {
      BLED = R[12] & 3;
 
      if ( (R[6] & 248) == 0 ) 		/* Group 0A - Basic tuning and switching info */
      {
        if ( BLED >= 0 )
        {
          pos = R[7] & 3; 		/* PS address 		*/
          CodearrayPS[pos*2] = R[10]; 	/* high byte 		*/
          CodearrayPS[pos*2+1] = R[11]; /* low byte 		*/
          altFreq = R[8];		/* 1. alternative freq	*/
          altFreq2 = R[9]; 		/* 2. alternative freq	*/
        }

      /* 1st alternative frequency */
      if ( altFreq == 250 ) { AM = 1; }		/* AM freq follows */
      if ( altFreq == 224 ) { noAFs = 1; }	/* No AFs follow !! */

      /* total number of following AFs */
      if ( ( altFreq > 224) && ( altFreq < 250 ) )
      {
        numFreq = altFreq - 224;
        altFreq = 0;
        // OUTPUT_DEBUG_STRING( sprintf("AFs to come: %d\n", numFreq) );
      }
           
      if ( AM == 0 )
      {
        if ( (altFreq > 0) && (altFreq <  205) )
        {
          altFreq = 8750 + altFreq * 10;
        }
      }

      if ( AM == 1 )
      {
        if ( (altFreq > 0) && (altFreq < 16) )
        {
          altFreq = 144 + altFreq * 9;
        }

        if ( (altFreq > 15) && (altFreq < 136) )
        {
          altFreq = 522 + altFreq * 9;
        }
            
        AM = 0;
      }

      /* 2nd alternative frequency */
      if ( altFreq2 == 250 ) { AM = 1; }
 
      if ( AM == 0 )
      {
        if ( (altFreq2 > 0) && (altFreq2 <  205) )
        {
          altFreq2 = 8750 + altFreq2 * 10;
        }
      }

      if ( AM == 1 )
      {
        if ( (altFreq2 > 0) && (altFreq2 < 16) )
        {
          altFreq2 = 144 + altFreq2 * 9;
        }

        if ( (altFreq2 > 15) && (altFreq2 < 136) )
        {
          altFreq2 = 522 + altFreq2 * 9;
        }
            
        AM = 0;
      }
          
      /* TP - Traffic Programme Identification bit (&4) */
      TP = (R[6] & 4) / 4;

      /* TA - Traffic Programme Identification bit (&16) */
      TA = (R[7] & 16) / 16;

      /* M/S - Music / Speech "Switch" bit (&8); 0 = speech | 1 = music or not used */
      MS = (R[7] & 8) / 8;
          
      /* PTY - Programm Type Code (&31) */
      int PTYtmp = ((256 * R[6] + R[7]) >> 5) & 0x1F;
      PTY = PTYtmp;
    } 
    else if ( (R[6] & 248) == 8 ) 		/* Group 0B - Basic tuning and switching info */
    {
      if ( BLED >= 0 )
      {
        pos = R[7] & 3; 		/* PS address 		*/
        CodearrayPS[pos*2] = R[10]; 	/* high byte 		*/
        CodearrayPS[pos*2+1] = R[11]; /* low byte 		*/
        /* int PIcode2 = 256 * R[8] + R[9]; */ /* alternative PI code */
      }

      /* TP - Traffic Programme Identification bit (&4) */
      TP = (R[6] & 4) / 4;

      /* TA - Traffic Programme Identification bit (&16) */
      TA = (R[7] & 16) / 16;

      /* M/S - Music / Speech "Switch" bit (&8); 0 = speech | 1 = music or not used */
      MS = (R[7] & 8) / 8;
          
      /* PTY - Programm Type Code (&31) */
      int PTYtmp = ((256 * R[6] + R[7]) >> 5) & 0x1F;
      PTY = PTYtmp;
    } 
    else if ( (R[6] & 248) == 16 ) 		/* Group 1A - Program Item Number (PIN) and Slow Labeling Codes */
    {
      variant = (R[8] & 128);

      int SLC_LA = R[8] & 128 >> 8;
      int SLC_VA = R[8] & 112 >> 5;
      
      int PIN_DAY = R[10] & 248 >> 3;
      int PIN_HH  = ( (R[10] & 7) << 2) + ( (R[11] & 192) >> 5);
      int PIN_MM  = R[11] & 63;
            
      printf(" Group 1A - Variant %d\n", variant);
      printf(" Group 1A - PIN: %d %d:%d\n", PIN_DAY, PIN_HH, PIN_MM);
      printf(" Group 1A - SLC LA: %d\n", SLC_LA);
      printf(" Group 1A - SLC VA: %d\n", SLC_VA);

    } 
    else if ( (R[6] & 248) == 32 ) 	/* Group 2A = Radiotext */
    {
      // OUTPUT_DEBUG_STRING( sprintf("Radiotext n: %d\n", n) );
    
      pos =  R[7] & 15;

      RTflagAB = (R[7] & 16);
      // OUTPUT_DEBUG_STRING( sprintf("RTflagAB: %d\n", RTflagAB ));

      if ( (pos == 0) && (RTflagAB == 0) )
      {
        if ( (R[8] != Codearray_RT_A[0]) && (Codearray_RT_A[0] != 0x20) )
        {
          for ( i = 0; i <= 63; i++ )
          {
            Codearray_RT_A[i] = 0x20;
          }
        }
      }

      if ( (pos == 0) && (RTflagAB == 16) )
      {
        if ( (R[8] != Codearray_RT_B[0]) && (Codearray_RT_B[0] != 0x20) )
        {
          for ( i = 0; i <= 63; i++ )
          {
            Codearray_RT_B[i] = 0x20;
          }
        }
      }

      if ( RTflagAB == 0 )
      {
        Codearray_RT_A[pos*4] = R[8];
        Codearray_RT_A[pos*4+1] = R[9];
        Codearray_RT_A[pos*4+2] = R[10];
        Codearray_RT_A[pos*4+3] = R[11];
      }
      else
      {
        Codearray_RT_B[pos*4] = R[8];
        Codearray_RT_B[pos*4+1] = R[9];
         Codearray_RT_B[pos*4+2] = R[10];
        Codearray_RT_B[pos*4+3] = R[11];
      }
    }
    else if ( (R[6] & 248) == 40 ) 	/* Group 2B = Radiotext */
    {
      // OUTPUT_DEBUG_STRING( sprintf("Radiotext n: %d\n", n) );
      pos =  R[7] & 15;

      RTflagAB = (R[7] & 16);
      // OUTPUT_DEBUG_STRING( sprintf("RTflagAB: %d\n", RTflagAB ));

      if ( (pos == 0) && (RTflagAB == 0) )
      {
        if ( (R[8] != Codearray_RT_A[0]) && (Codearray_RT_A[0] != 0x20) )
        {
          for ( i = 0; i <= 63; i++ )
          {
            Codearray_RT_A[i] = 0x20;
          }
        }
      }

      if ( (pos == 0) && (RTflagAB == 16) )
      {
        if ( (R[8] != Codearray_RT_B[0]) && (Codearray_RT_B[0] != 0x20) )
        {
          for ( i = 0; i <= 63; i++ )
          {
            Codearray_RT_B[i] = 0x20;
          }
        }
      }

      if ( RTflagAB == 0 )
      {
        Codearray_RT_A[pos*4] = R[8];
        Codearray_RT_A[pos*4+1] = R[9];
        Codearray_RT_A[pos*4+2] = R[10];
        Codearray_RT_A[pos*4+3] = R[11];
      }
      else
      {
        Codearray_RT_B[pos*4] = R[8];
        Codearray_RT_B[pos*4+1] = R[9];
         Codearray_RT_B[pos*4+2] = R[10];
        Codearray_RT_B[pos*4+3] = R[11];
      }
    }
    else if ( (R[6] & 248) == 48 ) 		/* Group 3A - Application Identification for ODA */
    {
      int AGTC = 0;
      int CBflag = 0;
      
      AGTC = R[7] & 31;
          
      int AID = R[10] * 256 + R[11];

      char str[256];
      sprintf(str, "Group 3A: %d%c %d (0x%x)\n", AGTC/2, ((AGTC & 1) == 0) ? 'A' : 'B', AID, AID);

      if ( dbgFlag.RDS )
        printf("3A: %s\n", str);

      /* RTplus */
      if ( AID == 0x4BD7 )
      {
        CBflag = R[9];  
        if ( dbgFlag.RDS )
          printf("3A - RTplus: CBflag = %d / %x\n", CBflag, CBflag);
      }
    }
    else if ( (R[6] & 248) == 64 ) /* Group 4A - Time and Date */
    {
      UTCoffset = (R[11] & 31) / 2;

      if ( (R[11] & 32) == 32 )
      {
        UTCoffset = UTCoffset * -1;
      }

      Minuten = R[11] / 64 + 4 * (R[10] & 15);
      Stunden = R[10] / 16 + 16 * (R[9] & 1);
      Zeit = 1;

      /* adjust local hrs for UTC+n */ 
      int StundenLocal = Stunden + UTCoffset;
      if ( StundenLocal > 23 )
        StundenLocal = StundenLocal - 24;

      sprintf(time_str, "%02d:%02d / %02d:%02d", Stunden, Minuten, StundenLocal, Minuten);

      // OUTPUT_DEBUG_STRING( sprintf("RDS - 4A - Time: %s\n", time_str) );

      /* update GUI text entry field here, only if new time group found */
      gtk_entry_set_text(GTK_ENTRY(entry_rds_time), (char *)time_str);
    }
    else if ( ((R[6] >> 4) == 8) && ((R[6] & 8) == 0) ) /* Group 8A - TMC */
    {
      if ( dbgFlag.TMC )
      {
        OUTPUT_DEBUG_STRING( "RDS - 8A - possible TMC data found!\n" );

        /* T - user message? */
        TMC_T = R[7] & 16;

        /* Single or Multi Group Message? */
        TMC_F = R[7] & 8;

        /* valid? */
        TMC_DP = R[7] & 7;
      
        /* Umfahrung noetig? / redirection? */
        TMC_D = R[8] & 128;
      
        /* direction (positive or negative) */
        TMC_R = R[8] & 64;
      
        /* extent of event in loc codes */
        TMC_Extent = R[8] & 56;

        /* Event code */
        TMC_Event = (R[8] & 7) * 256 + R[9];

        /* Location code */
        TMC_Loc = R[10] * 256 + R[11];
      
        printf("TMC user message  : %d\n", TMC_T);
        printf("TMC Single / Multi: %d\n", TMC_F);
        printf("TMC valid?        : %d\n", TMC_DP);
        printf("TMC redirection?  : %d\n", TMC_D);
        printf("TMC direction     : %d\n", TMC_R);
        printf("TMC extent        : %d\n", TMC_Extent);
        printf("TMC event code    : %d\n", TMC_Event);
        printf("TMC location code : %d\n\n", TMC_Loc);
      }
    }
    else if ( (R[6] & 248) == 160 ) 		/* Group 10A - PTYN */
    {
      pos = R[7] & 1; 	
      CodearrayPTYN[pos*4] = R[8];
      CodearrayPTYN[pos*4+1] = R[9];
      CodearrayPTYN[pos*4+2] = R[10];
      CodearrayPTYN[pos*4+3] = R[11];

      printf("PTYN: %s\n", (char *)CodearrayPTYN);
    }
    else if ( ((R[6] >> 4) == 12) && ((R[6] & 8) == 0) ) /* Group 12A - RT+ */
    {
      if ( dbgFlag.RDS )
        printf("12A: %x %x %x %x %x %x %x %x\n", R[4], R[5], R[6], R[7], R[8], R[9], R[10], R[11]);

      /* decode toogle and running bit */
      int toggle_bit  = ( (R[7] & 16) == 16 ) ? 1 : 0;
      int running_bit = ( (R[7] & 8) == 8 ) ? 1 : 0;

      RTplus1.toggle  = toggle_bit;
      RTplus2.toggle  = toggle_bit;
      RTplus1.running = running_bit;
      RTplus2.running = running_bit;

      /* get tag type */
      RTplus1.type = ((R[7] & 7) << 3) + ((R[8] & 224) >> 5);
      RTplus2.type = ((R[9] & 1) << 5) + ((R[10] & 248) >> 3);

      /* get start and length for RT+ tag #1 */
      RTplus1.start  = ((R[8] & 31) << 1) + ((R[9] & 128) >> 7);
      RTplus1.length = ((R[9] & 126) >> 1);

      /* get start and length for RT+ tag #2 */
      RTplus2.start  = ((R[10] & 7) << 3) + ((R[11] & 224) >> 5);
      RTplus2.length = (R[11] & 31);

      /* set counter */
      RTplus1.cnt = 2;
      RTplus2.cnt = 2;
    }
    else if ( ((R[6] >> 4) == 14) && ((R[6] & 8) == 0) ) /* Group 14A - EON */
    {
      if ( dbgFlag.RDS )
        printf("14A: %x %x %x %x %x %x %x %x\n", R[4], R[5], R[6], R[7], R[8], R[9], R[10], R[11]);
    }
  }      
  
    int Group1, Group2;
    Group1 = R[6] >> 4;
    Group2 = R[6] & 8;
    //OUTPUT_DEBUG_STRING( sprintf("Group: %d%s\n", Group1, ( Group1 == 0 ) ? "A" : "B") );
    // printf("Group: %d%s\n", Group1, ( Group2 == 0 ) ? "A" : "B");

    // GTcnt[GTA]++; 
  }

  if ( RTplus1.type > 0 )
  {
    if ( RTplus1.cnt > 0 )
      RTplus1.cnt--;
    
    if ( RTplus1.cnt == 1 )
    {
      if (  RTplus1.ab_flag == 0 )
      {
        if ( (RTplus1.type > 0) && (strlen((char *)Codearray_RT_A) >= (RTplus1.start + RTplus1.length)) )
          RTplus1.str = substring((const char *)Codearray_RT_A, RTplus1.start, RTplus1.length);
      }
      else if (  RTplus1.ab_flag == 1 )
      {
        if ( (RTplus1.type > 0) && (strlen((char *)Codearray_RT_B) >= (RTplus1.start + RTplus1.length)) )
          RTplus1.str = substring((const char *)Codearray_RT_B, RTplus1.start, RTplus1.length);
      }

      if ( dbgFlag.RTplus )
      {
        printf("RT+ Tag1 of type %2d (%-20s), starting at %d, length %d: %s\n", 
                  RTplus1.type, 
                  (char *)RTplus_Codes[RTplus1.type].rtp_class, 
                  RTplus1.start, 
                  RTplus1.length,
                  (char *)RTplus1.str);
      }
    }
    else if ( RTplus1.cnt == 15 )
    {
      /* store Radiotext AB Flag */
      RTplus1.ab_flag =  RTflagAB;
    }
  }

  if ( RTplus2.type > 0 )
  {
    if ( RTplus2.cnt > 0 )
      RTplus2.cnt--;
    
    if ( RTplus2.cnt == 1 )
    {
      if (  RTplus2.ab_flag == 0 )
      {
        if ( (RTplus2.type > 0) && (strlen((char *)Codearray_RT_A) >= (RTplus2.start + RTplus2.length)) )
          RTplus2.str = substring((const char *)Codearray_RT_A, RTplus2.start, RTplus2.length);
      }
      else if (  RTplus2.ab_flag == 1 )
      {
        if ( (RTplus2.type > 0) && (strlen((char *)Codearray_RT_B) >= (RTplus2.start + RTplus2.length)) )
          RTplus2.str = substring((const char *)Codearray_RT_B, RTplus2.start, RTplus2.length);
      }

      if ( dbgFlag.RTplus )
      {
        printf("RT+ Tag1 of type %2d (%-20s), starting at %d, length %d: %s\n", 
                  RTplus2.type, 
                  (char *)RTplus_Codes[RTplus2.type].rtp_class, 
                  RTplus2.start, 
                  RTplus2.length,
                  (char *)RTplus2.str);
      }
    }
    else if ( RTplus2.cnt == 15 )
    {
      /* store Radiotext AB Flag */
      RTplus2.ab_flag =  RTflagAB;
    }
  }

  /* *******************************************
     *** update display of RDS values in GUI *** 
     *******************************************/
  if ( R[3] >= 0 )
  {
    /* PI code */
    int rds_pi = 256 * R[4] + R[5];
    if ( rds_pi > 0 )
      RDS_PI = rds_pi;
    char pi_str[7];
    sprintf(pi_str, "0x%X", RDS_PI);

    /* AF - alternate frequencies */
    char af_str[20];

    if ( noAFs == 1 )
    {
      sprintf(af_str, "- keine AF -");
    }
    else if ( (altFreq > 0) && (altFreq2 > 0) )
    {
      sprintf(af_str, "%.2f / %.2f", (double) altFreq /100, (double) altFreq2 / 100);
    }
    else if ( (altFreq > 0) && (altFreq2 == 0) )
    {
      sprintf(af_str, "%.2f", (double) altFreq / 100);
    }
    else if ( (altFreq == 0) && (altFreq2 > 0) )
    {
      sprintf(af_str, "%.2f", (double) altFreq2 / 100);
    }
    else
    {
      sprintf(af_str, "%s", "-- AF --");
    }

    /* PTY - programme type */
    char pty_str[13];
    sprintf(pty_str, "%02d %s", PTY, (char *)PTY_Table_Short[PTY]);

    gtk_entry_set_text(GTK_ENTRY(entry_rds_af), (char *)af_str);
    gtk_entry_set_text(GTK_ENTRY(entry_rds_pi), (char *)pi_str);
    gtk_entry_set_text(GTK_ENTRY(entry_rds_pty), (char *)pty_str);
    gtk_entry_set_text(GTK_ENTRY(entry_rds_psn), (char *)CodearrayPS);

    /* fix RadioText strings */
    int fixA = 0, fixB = 0;
    for ( i = 0; i <= 63; i++ )
    {
      if ( Codearray_RT_A[i] == 0x0D )
        fixA = 1;

      if ( Codearray_RT_B[i] == 0x0D )
        fixB = 1;

      if ( fixA )
        Codearray_RT_A[i] = 0x20;

      if ( fixB )
        Codearray_RT_B[i] = 0x20;
    }
  
    /* convert radiotext A string to UTF-8 for display with Gtk ?? */
    //RT_utf8 = g_convert ( (gchar*)Codearray_RT_A, sizeof(Codearray_RT_A), "ISO-8859-15", "UTF-8", 
    //                           NULL, NULL, NULL);

    unsigned char *RT_utf8_A = rds_charset_decode((unsigned char*)Codearray_RT_A);                          
    gtk_entry_set_text(GTK_ENTRY(entry_rds_rt_a), (gchar *)RT_utf8_A);
    free(RT_utf8_A);
  
    /* convert radiotext B string to UTF-8 for display with Gtk ?? */
    //RT_utf8 = g_convert ( (gchar*)Codearray_RT_B, sizeof(Codearray_RT_B), "ISO-8859-1", "UTF-8", 
    //                           NULL, NULL, NULL);

    unsigned char *RT_utf8_B = rds_charset_decode((unsigned char*)Codearray_RT_B);                          
    gtk_entry_set_text(GTK_ENTRY(entry_rds_rt_b), (gchar *)RT_utf8_B);
    free(RT_utf8_B);
  
    /* update indicators for M/S, TP and TA bit settings */
    ( MS ) ? gtk_widget_modify_base(entry_rds_ms, GTK_STATE_NORMAL, &colorGreen) : gtk_widget_modify_base(entry_rds_ms, GTK_STATE_NORMAL, &colorWhite);
    ( TP ) ? gtk_widget_modify_base(entry_rds_tp, GTK_STATE_NORMAL, &colorGreen) : gtk_widget_modify_base(entry_rds_tp, GTK_STATE_NORMAL, &colorWhite);
    ( TA ) ? gtk_widget_modify_base(entry_rds_ta, GTK_STATE_NORMAL, &colorGreen) : gtk_widget_modify_base(entry_rds_ta, GTK_STATE_NORMAL, &colorWhite);
  }
}

void RDS_Get_PSN(char *tmpPS)
{
  int n, pos, BLED;
  char PS[32];
  PS[0] = '\0';
  int BM = 0;
  
  /* vars for group 1A */
  int altFreq, altFreq2, numFreq, AM;
  altFreq = 0; altFreq2 = 0;
  numFreq = 0;
  AM = 0;

  for ( n = 1; n <= 64; n++ )
  {
    while ( gtk_events_pending() )
    {
      gtk_main_iteration();
    }

    if ( GET_INT_STATUS() )
    {
      FM_RDS_STATUS();
    }
    
    if ( R[3] > 0 )
    {
      BLED = R[12] & 3;
 
      if ( (R[6] & 248) == 0 ) 		/* Group 0A - Basic tuning and switching info */
      {
        if ( BLED >= 0 )
        {
          pos = R[7] & 3; 	/* PS address 		*/
          PS[pos*2] = R[10]; 	/* high byte 		*/
          PS[pos*2+1] = R[11]; 	/* low byte 		*/
          altFreq = R[8];	/* 1. alternative freq	*/
          altFreq2 = R[9]; 	/* 2. alternative freq	*/

          /* set bitmask to determine if pSN is complete (=127) */
          BM |= Bitmask[pos * 2];
          BM |= Bitmask[pos * 2 + 1];
        }

        /* 1st alternative frequency */
        if ( altFreq == 250 ) { AM = 1; }	/* AM freq follows */
 
        /* total number of following AFs */
        if ( ( altFreq > 224) && ( altFreq < 250 ) )
        {
          numFreq = altFreq - 224;
          altFreq = 0;
        }
           
        if ( AM == 0 )
        {
          if ( (altFreq > 0) && (altFreq <  205) )
          {
            altFreq = 8750 + altFreq * 10;
          }
        }

        if ( AM == 1 )
        {
          if ( (altFreq > 0) && (altFreq < 16) )
          {
            altFreq = 144 + altFreq * 9;
          }

          if ( (altFreq > 15) && (altFreq < 136) )
          {
            altFreq = 522 + altFreq * 9;
          }
            
          AM = 0;
        }

        /* 2nd alternative frequency */
        if ( altFreq2 == 250 ) { AM = 1; }
 
        if ( AM == 0 )
        {
          if ( (altFreq2 > 0) && (altFreq2 <  205) )
          {
            altFreq2 = 8750 + altFreq2 * 10;
          }
        }

        if ( AM == 1 )
        {
          if ( (altFreq2 > 0) && (altFreq2 < 16) )
          {
            altFreq2 = 144 + altFreq2 * 9;
          }

          if ( (altFreq2 > 15) && (altFreq2 < 136) )
          {
            altFreq2 = 522 + altFreq2 * 9;
          }
            
          AM = 0;
        }
          
        /* TP - Traffic Programme Identification bit (&4) */
        TP = (R[6] & 4) / 4;

        /* TA - Traffic Programme Identification bit (&16) */
        TA = (R[7] & 16) / 16;

        /* M/S - Music / Speech "Switch" bit (&8); 0 = speech | 1 = music or not used */
        MS = (R[7] & 8) / 8;

        /* PTY - Programm Type Code (&31) */
        int PTYtmp = ((256 * R[6] + R[7]) >> 5) & 0x1F;
        PTY = PTYtmp;
      } 
      else if ( (R[6] & 248) == 8 ) 		/* Group 0B - Basic tuning and switching info */
      {
        if ( BLED >= 0 )
        {
          pos = R[7] & 3; 	/* PS address 		*/
          PS[pos*2] = R[10]; 	/* high byte 		*/
          PS[pos*2+1] = R[11]; 	/* low byte 		*/

          /* set bitmask to determine if pSN is complete (=127) */
          BM |= Bitmask[pos * 2];
          BM |= Bitmask[pos * 2 + 1];
        }

        /* TP - Traffic Programme Identification bit (&4) */
        TP = (R[6] & 4) / 4;

        /* TA - Traffic Programme Identification bit (&16) */
        TA = (R[7] & 16) / 16;

        /* M/S - Music / Speech "Switch" bit (&8); 0 = speech | 1 = music or not used */
        MS = (R[7] & 8) / 8;
          
        /* PTY - Programm Type Code (&31) */
        int PTYtmp = ((256 * R[6] + R[7]) >> 5) & 0x1F;
        PTY = PTYtmp;
      } 

      if ( BM == 127 )
      {
        break;
      }
    }
    else
    {
      /* no RDS info? */
      break;
    }      
  }

  int rds_pi = 256 * R[4] + R[5];
  char pi_str[7];
  sprintf(pi_str, "0x%X", rds_pi);

  strcpy(tmpPS, PS);
}


/* ============================
   .::. GTK+ WINDOW PART 1 .::. 
   ============================ */

static gboolean time_handler(GtkWidget *widget)
{
  if (widget->window == NULL) return FALSE;

  // gtk_widget_queue_draw(widget);

  RDSupdate();

  return TRUE;
}

/* ============================
   .::. Start FM/AM Mode .::. 
   ============================ */

void FM_Start(int f, int vol)
{
  int i;
  int fm_arr_size = sizeof( SI4735_prop_fm )/sizeof(  SI4735_prop_fm[0]);

  PowerDown();
  Delay(50);

  Init_FM();
  Delay(200);
     
  /* get ic info */
  GetRev();

  sprintf(info_si.ic, "47%d", R[1]);
  sprintf(info_si.fw, "%c%c", R[2], R[3]);
  sprintf(info_si.comp_fm, "%c%c", R[6], R[7]);
  sprintf(info_si.chip_rev, "%c", R[8]);

  Delay(110);

  /* set FM properties */
  // Prop(0x1107, 0x0000); /* FM_ANTENNA_INPUT */
  // Prop(0x1100, 0x0001); /* FM Deemphasis */
  // Prop(0x1402, 0x0005); /* Freq. Spacing 5 kHz */
  for ( i = 0; i < fm_arr_size; i++ )
  {
    if ( SI4735_prop_fm[i].val_curr > -1 )
    {
      Prop(SI4735_prop_fm[i].prop, SI4735_prop_fm[i].val_curr);
      Delay(110);
    }
  }
  
  RX_VOLUME(vol);	/* set initial volume */

  Delay(110);
     
  Init_RDS();		/* initialize RDS */

  Delay(110);

  FM_Tune_Freq(f);	/* tune initial frequency */

  Delay(110);

  if ( GET_INT_STATUS() )
  {
    FM_Tune_Status(3);
  }
 
  isModeFM = 1;
  isModeAM = 0;
  
  char *gstr = "FM Mode running ....\0";
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), gstr), gstr);
}

void AM_Start(int f, int vol)
{
  int i;
  int am_arr_size = sizeof( SI4735_prop_am )/sizeof(  SI4735_prop_am[0]);

  GPOstate = 0;

  PowerDown();
  Delay(50);
  Init_AM();
  Delay(110);
     
  /* get ic info */
  GetRev();

  sprintf(info_si.ic, "47%d", R[1]);
  sprintf(info_si.fw, "%c%c", R[2], R[3]);
  sprintf(info_si.comp_am, "%c%c", R[6], R[7]);
  sprintf(info_si.chip_rev, "%c", R[8]);

  Delay(110);

  /* set AM properties from application settings / stored */
  for ( i = 0; i < am_arr_size; i++ )
  {
    if ( SI4735_prop_am[i].val_curr > -1 )
    {
      Prop(SI4735_prop_am[i].prop, SI4735_prop_am[i].val_curr);
      Delay(110);
    }
  }
     
  RX_VOLUME(vol);

  Delay(110);
     
  AM_Tune_Freq(f);

  Delay(110);

  if ( GET_INT_STATUS() )
  {
    AM_Tune_Status(3);
  }

  /* band borders: search start freq at begin of band bottom */
  for ( i = 0; i <= 14; i++ )
  {
    if ( ( f >=  AM_bands[i].band_bottom ) && ( f <=  AM_bands[i].band_top ) )
    {
      /* set prop: frequency spacing */
      Prop( 0x3402, AM_bands[i].freq_spacing);
      Delay(100);
    }
  }

  /* set raio mode state */
  isModeFM = 0;
  isModeAM = 1;    

  char *gstr = "AM Mode running ....\0";
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), gstr), gstr);

}

/* ============================
   .::. GTK+ WINDOW PART 2 .::. 
   ============================ */

void ende_cb(GtkObject *object)
{
  /* save presets */
  si4735radio_presets_save(); 

  /* save settings */
  si4735radio_settings_save();

  /* close FT Handle */
#ifdef FTD2XX
  FT_Close(ftHandle);
#else
  ftdi_usb_close(&ftdic);
  ftdi_deinit(&ftdic);
#endif

  /* quit Gtk main window */
  gtk_main_quit();
}

void exit_cb(GtkObject *object)
{
  /* save presets */
  si4735radio_presets_save(); 
  
  /* save settings */
  si4735radio_settings_save();

  /* leave application and power down radio before */
  PowerDown();
  Delay(50);

  /* close FT Handle */
#ifdef FTD2XX
  FT_Close(ftHandle);
#else
    ftdi_usb_close(&ftdic);
    ftdi_deinit(&ftdic);
#endif

  /* quit Gtk main window */
  gtk_main_quit();
}

/* ---------------------------------
    callback routines: FM radio
   --------------------------------- */

void fm_start_cb(GtkObject *object)
{
  FM_Start(startFreqFM, startVolume);

  /* auto start RDS decoding */
  if ( startRDS )
  {
    if ( rdsaktiv == 1 )
    {
      g_source_remove( rdsTimer );
      rdsaktiv = 0;
      gtk_button_set_label(GTK_BUTTON(button_rds), "RDS starten");
    }
    else
    {
      rdsTimer = g_timeout_add(5, (GSourceFunc) time_handler, (gpointer) object);
      rdsaktiv = 1;
      gtk_button_set_label(GTK_BUTTON(button_rds), "RDS stoppen");
    }
  }
}

void fm_seek_cb(GtkObject *object, gpointer user_data )
{
  int n, arg;

  if ( strcmp( (char *)user_data, "seek down" ) == 0)
  {
    arg = 0; // seek down, halt at band end (no wrap)
  }
  else
  {
    arg = 8; // seek up, halt at band end (no wrap)
  }

  FM_Seek_Freq(arg);
  
  sleep(1);
  
  if ( GET_INT_STATUS() )
  {
    FM_Tune_Status(3);
  }
  
  FM_SMeter();

  for ( n = 0; n <= 63; n++ ) { Codearray_RT_A[n] = 32; Codearray_RT_B[n] = 32;}
  for ( n = 0; n <= 63; n++ ) { CodearrayRT2[n] = 32; }
  for ( n = 0; n <= 32; n++ ) { CodearrayPS[n] = 32; tmpCodearrayPS[n] = 32; }

  RDS_PI = 0; TP = 0; TA = 0; MS = 0; PTY = 0; MS = 0;
}

void fm_autoscan_cb(GtkObject *object )
{
  /* declare some vars */
  int f, f_tuned, n, m, oldFreq;
  int p_start = 0;
  int free_only = 0;
  
  /* ask user: overwrite presets or append? */
  GtkWidget *dlg;
  dlg = gtk_message_dialog_new (GTK_WINDOW(window),
                        	   GTK_DIALOG_DESTROY_WITH_PARENT,
                                   GTK_MESSAGE_QUESTION,
                                   GTK_BUTTONS_YES_NO,
                                   "Ja - Speicherplätze löschen und neu belegen\nNein - Gefundene Sender auf freie Speicherplätze legen");

  int res = gtk_dialog_run (GTK_DIALOG (dlg));
  gtk_widget_destroy (dlg);

  printf("Dialog ResCode: %d\n", res);

  if ( res == GTK_RESPONSE_YES )
  {
    for ( m = 0; m < NUM_FM_PRESETS; m++ )
    {  
      FM_Preset[m].freq = 0;
      strcpy(FM_Preset[m].station_name, "");
    }
  }
  else // if ( res == GTK_RESPONSE_NO )
  {
    free_only = 1;
  }
  
  char *gstr = "FM Mode running .... AutoScan läuft ....\0";
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), gstr), gstr);

  /* save olf frequency (currently tuned) */
  if ( GET_INT_STATUS() )
  {
    FM_Tune_Status(3);
  }

  f_tuned = (R[2] * 256 + R[3]);
  oldFreq = f_tuned;

  /* begin at lower band end */
  f = FM_Band_Bottom[ FM_Band_Selected ];
  waitCTS();
  FM_Tune_Freq(f);
  Delay (110);

  for ( m = p_start; m < NUM_FM_PRESETS; m++ )
  {  
    while ( gtk_events_pending() )
    {
      gtk_main_iteration();
    }

    if ( GET_INT_STATUS() )
    {
      FM_Seek_Freq(8);
    }
      
    sleep(1);
  
    if ( GET_INT_STATUS() )
    {
      FM_Tune_Status(0);
    }

    sleep(1);

    f_tuned = (R[2] * 256 + R[3]);

    /* get RDS name */
    char tmpPS[32];
    tmpPS[0] = '\0';
    RDS_Get_PSN( tmpPS );

    if ( f_tuned >= (FM_Band_Top[ FM_Band_Selected ] - 10) )
      break;
    
    /* search free / unused preset */
    if ( free_only )
    {
      while ( ( FM_Preset[m].freq > 8750 ) && ( m < NUM_FM_PRESETS) )
      {
        m++;
      }
    }
    
    if ( m == NUM_FM_PRESETS )
    { 
      /* all presets already taken - inform user */
      dlg = gtk_message_dialog_new (GTK_WINDOW(window),
                              	     GTK_DIALOG_DESTROY_WITH_PARENT,
                              	     GTK_MESSAGE_WARNING,
                              	     GTK_BUTTONS_OK,
                              	     "Alle Speicherplätze belegt!");

      res = gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);

      break;
    }
    
    FM_Preset[m].freq = f_tuned;
    strcpy(FM_Preset[m].station_name, tmpPS);
     
    if ( m < 6 )
    {
      char pLabel[32];

      if ( strlen(FM_Preset[ m ].station_name) > 2 )
      {
        sprintf(pLabel, "%s", FM_Preset[ m ].station_name );
      }
      else
      {
        sprintf(pLabel, "%.2f MHz", (double)f_tuned/100 );
      }

      gtk_button_set_label(GTK_BUTTON(button_fm_preset[m]), pLabel);    
    }
    
    for ( n = 0; n <= 63; n++ ) { Codearray_RT_A[n] = 32; Codearray_RT_B[n] = 32;}
    for ( n = 0; n <= 63; n++ ) { CodearrayRT2[n] = 32; }
    for ( n = 0; n <= 32; n++ ) { CodearrayPS[n] = 32; tmpCodearrayPS[n] = 32; }
  }

  /* restore freq tuned before autoscan */
  FM_Tune_Freq(oldFreq);

  gstr = "FM Mode running .... AutoScan beendet.\0";
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), gstr), gstr);
}

void fm_tune_preset_cb(GtkObject *object, gpointer user_data )
{
  int n;

  int preset = (int)user_data;
  int freq = FM_Preset[ FM_Preset_Bank * 6 + preset ].freq;
  
  if ( (freq > FM_Band_Bottom[ FM_Band_Selected ]) && (freq < FM_Band_Top[ FM_Band_Selected ]) )
  {
    waitCTS();
    FM_Tune_Freq(freq);
    Delay (110);
  
    if ( GET_INT_STATUS() )
    {
      FM_Tune_Status(3);
    }

    FM_SMeter();

    for ( n = 0; n <= 63; n++ ) { Codearray_RT_A[n] = 32; Codearray_RT_B[n] = 32;}
    for ( n = 0; n <= 63; n++ ) { CodearrayRT2[n] = 32; }
    for ( n = 0; n <= 32; n++ ) { CodearrayPS[n] = 32; tmpCodearrayPS[n] = 32; }

    RDS_PI = 0; TP = 0; TA = 0; MS = 0; PTY = 0; MS = 0;
  }
}

gboolean fm_preset_onButtonPressed (GtkObject *object, GdkEventButton *event, gpointer user_data)
{
  int preset = (int)user_data;

  /* single click with the right mouse button? */
  if (event->type == GDK_BUTTON_PRESS && event->button == 3)
  {
    // printf("Single right click on FM preset %d.\n", preset);

    if ( GET_INT_STATUS() )
    {
      FM_Tune_Status(3);
    }

    sleep(1);

    int f_tuned = (R[2] * 256 + R[3]);

    /* get RDS name */
    char tmpPS[32];
    tmpPS[0] = '\0';
    RDS_Get_PSN( tmpPS );

    FM_Preset[FM_Preset_Bank * 6 + preset].freq = f_tuned;
    strcpy( FM_Preset[FM_Preset_Bank * 6 + preset].station_name, tmpPS);

    if ( preset < 6 )
    {
      char pLabel[20];
      sprintf(pLabel, "%.2f MHz", (double)f_tuned/100 );
      gtk_button_set_label(GTK_BUTTON(button_fm_preset[preset]), pLabel);    
    }

    return TRUE;
  }
  /* single click with the right mouse button? */
  else if (event->type == GDK_BUTTON_PRESS && event->button == 1)
  {
    // printf("Single left click on FM preset %d.\n", preset);

    int n;
    int freq = FM_Preset[ FM_Preset_Bank * 6 + preset ].freq;

    if ( (freq > FM_Band_Bottom[ FM_Band_Selected ]) && (freq < FM_Band_Top[ FM_Band_Selected ]) )
    {
      waitCTS();
      FM_Tune_Freq(freq);
      Delay (110);
  
      if ( GET_INT_STATUS() )
      {
        FM_Tune_Status(3);
      }

      FM_SMeter();

      for ( n = 0; n <= 63; n++ ) { Codearray_RT_A[n] = 32; Codearray_RT_B[n] = 32;}
      for ( n = 0; n <= 63; n++ ) { CodearrayRT2[n] = 32; }
      for ( n = 0; n <= 32; n++ ) { CodearrayPS[n] = 32; tmpCodearrayPS[n] = 32; }

      RDS_PI = 0; TP = 0; TA = 0; MS = 0; PTY = 0; MS = 0;
    }
    return TRUE;
  }
  else if (event->type == GDK_BUTTON_PRESS && event->button == 2)
  {
    // printf("Middle click on FM preset %d.\nDeleting preset.\n", preset);

    GtkWidget *dlg;
    dlg = gtk_message_dialog_new (GTK_WINDOW(window),
                              	   GTK_DIALOG_DESTROY_WITH_PARENT,
                                   GTK_MESSAGE_QUESTION,
                                   GTK_BUTTONS_YES_NO,
                                   "Speicherplatz löschen?");

    int res = gtk_dialog_run (GTK_DIALOG (dlg));
    gtk_widget_destroy (dlg);

    if ( res == GTK_RESPONSE_YES )
    {
      FM_Preset[FM_Preset_Bank * 6 + preset].freq = 0;
      strcpy( FM_Preset[FM_Preset_Bank * 6 + preset].station_name, "");

      if ( preset < 6 )
      {
        int pr = FM_Preset_Bank * 6 + preset + 1;
        char pLabel[20];
        sprintf(pLabel, " - %d - ", pr );
        gtk_button_set_label(GTK_BUTTON(button_fm_preset[preset]), pLabel);    
      }
    }

    return TRUE;
  }
  else if (event->type == GDK_BUTTON_PRESS && event->button == 4)
  {
    printf("Mouse Wheel Scroll Up on FM preset %d.\n", preset);
  }
  else if (event->type == GDK_BUTTON_PRESS && event->button == 5)
  {
    printf("Mouse Wheel Scroll Down on FM preset %d.\n", preset);
  }
  
  return FALSE;
}

void fm_preset_change_bank_cb(GtkObject *object )
{
  int f, m;
  int bank = FM_Preset_Bank + 1;
  int p_start = 6 * bank;
  char pLabel[32];
  char tLabel[32];
    
  if ( (p_start + 6) <= (sizeof(FM_Preset) / sizeof(FM_Preset[0])) )
  {
    FM_Preset_Bank = bank;
  }
  else
  {
    FM_Preset_Bank = 0;
  }

  /* label preset memory bank button */
  sprintf(pLabel, "MEM %d", (FM_Preset_Bank + 1) );
  gtk_button_set_label(GTK_BUTTON(button_fm_preset_bank), pLabel);    
 
  for ( m = 0; m < 6; m++ )
  {
    /* get frequency */
    f = FM_Preset[ FM_Preset_Bank * 6 + m ].freq;

    /* relabel preset button */
    if ( f > 0 )
    {
      if ( strlen(FM_Preset[ FM_Preset_Bank * 6 + m ].station_name) > 2 )
      {
        // sprintf(pLabel, "%d - %s", (FM_Preset_Bank * 6 + m+1), FM_Preset[ FM_Preset_Bank * 6 + m ].station_name );
        sprintf(pLabel, "%s", FM_Preset[ FM_Preset_Bank * 6 + m ].station_name );
      }
      else
      {
        // sprintf(pLabel, "%d - %.2f", (FM_Preset_Bank * 6 + m+1), (double)f/100 );
        sprintf(pLabel, "%.2f MHz", (double)f/100 );
      }
      
      sprintf(tLabel, "P %d - %.2f MHz - %s", (FM_Preset_Bank * 6 + m+1), (double)f/100, FM_Preset[ FM_Preset_Bank * 6 + m ].station_name );
      
      gtk_button_set_label(GTK_BUTTON(button_fm_preset[m]), pLabel);    

      gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_preset[m],
                               tLabel, "");
    }
    else
    {
      sprintf(pLabel, "- %d -", (FM_Preset_Bank * 6 + m+1));
      gtk_button_set_label(GTK_BUTTON(button_fm_preset[m]), pLabel);    
    }
  }  
}

void fm_rds_cb(GtkObject *object)
{
  if ( rdsaktiv == 1 )
  {
    g_source_remove( rdsTimer );
    rdsaktiv = 0;
    gtk_button_set_label(GTK_BUTTON(button_rds), "RDS starten");
  }
  else
  {
    rdsTimer = g_timeout_add(5, (GSourceFunc) time_handler, (gpointer) object);
    rdsaktiv = 1;
    gtk_button_set_label(GTK_BUTTON(button_rds), "RDS stoppen");
  }
}

void fm_smeter_update_cb(GtkObject *object)
{
  FM_SMeter();
}

void change_fm_digits( GtkWidget *widget,
                       GtkSpinButton *spin )
{
  double f = 0.0;

  gtk_spin_button_set_digits (GTK_SPIN_BUTTON (entry_fm_freq), 2);

  f = gtk_spin_button_get_value (spin);

  autoTune = 0;

  if ( autoTune == 0 )
  {
    int f2, n;
    f2 = (int) (f * 100);

    waitCTS();
    FM_Tune_Freq(f2);
    Delay (110);

    /* update SMeter */
    FM_SMeter();

    /* clear RDS strings */
    for ( n = 0; n <= 63; n++ ) { Codearray_RT_A[n] = 32; Codearray_RT_B[n] = 32; CodearrayRT2[n] = 32; }
    for ( n = 0; n <= 32; n++ ) { CodearrayPS[n] = 32; tmpCodearrayPS[n] = 32; }

    /* clear RDS flags */
    RDS_PI = 0; TP = 0; TA = 0; MS = 0; PTY = 0; MS = 0;    
  }
}

/* ---------------------------------
    callback routines: AM radio
   --------------------------------- */

void am_start_cb(GtkObject *object)
{
  AM_Start(startFreqAM, startVolume);
}

void am_seek_cb(GtkObject *object, gpointer user_data )
{
  int arg;

  if ( strcmp( (char *)user_data, "seek down" ) == 0)
  {
    arg = 2; // seek down, halt at band end (no wrap)
  }
  else
  {
    arg = 10; // seek up, halt at band end (no wrap)
  }

  AM_Seek_Freq(arg);
  
  int n = 0;
  
  while ( (! GET_INT_STATUS()) || ( n < 10 ) )
  {
    Delay(20);
    AM_Tune_Status(0);
    n++;

    if ( (R[1] & 1) == 1 )
      break;
  }
  
  Delay(110);

  /* set GPIO */
  // AM_GPIO_Set(f_new);
  
  /* update s meter readings */
  AM_SMeter();
}

void am_tune_khz_cb(GtkObject *object, gpointer user_data )
{
  int f_step;
  int f_tuned;
  int f_new;

  /* step */
  f_step = (int)user_data;
  
  /* save old frequency (curently tuned) */
  if ( GET_INT_STATUS() )
  {
    AM_Tune_Status(3);
  }

  f_tuned = (R[2] * 256 + R[3]);

  f_new = f_tuned + f_step;

  if ( f_new < 150 )
    f_new = 150;
    
  if (f_new > 23000)
    f_new = 23000;

  waitCTS();
  AM_Tune_Freq(f_new);
  Delay (110);
  
  if ( GET_INT_STATUS() )
  {
    AM_Tune_Status(0); // (3) ?
  }
  
  /* set GPIO */
  AM_GPIO_Set(f_new);
  
  /* update s meter readings */
  AM_SMeter();
}

void am_autoscan_cb(GtkObject *object )
{
  int f = 0;
  int f_tuned, m, oldFreq;
  int band = -1;
  int n;
  
  char *gstr = "AM Mode running .... AutoScan läuft ....\0";
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), gstr), gstr);
  
  /* save old frequency (curently tuned) */
  if ( GET_INT_STATUS() )
  {
    AM_Tune_Status(3);
  }

  f_tuned = (R[2] * 256 + R[3]);
  oldFreq = f_tuned;

  /* search start freq at begin of band bottom */
  for ( m = 0; m <= 14; m++ )
  {
    if ( ( oldFreq >=  AM_bands[m].band_bottom ) && ( oldFreq <=  AM_bands[m].band_top ) )
    {
      f = AM_bands[m].band_start_freq;
      band = m;

      /* set prop: frequency spacing */
      Prop( 0x3402, AM_bands[m].freq_spacing);
      Delay(100);
    }
  }
  
  waitCTS();
  AM_Tune_Freq(f);
  Delay (110);

  for ( m = 0; m < NUM_AM_PRESETS; m++ )
  {  
    while ( gtk_events_pending() )
    {
      gtk_main_iteration();
    }

    if ( GET_INT_STATUS() )
    {
      AM_Seek_Freq(8);
    }
      
    /* search, max 5 sek */
    n = 0;
    while ( n < 50 )
    { 
      n++;
      
      if ( GET_INT_STATUS() )
      {
        AM_Tune_Status(0);
      }

      Delay(100);

      if ( (R[1] & 1) == 1 )
        break;
    } 

    f_tuned = (R[2] * 256 + R[3]);

    if ( f_tuned >= (  AM_bands[band].band_top - 10) )
      break;
    
    AM_Preset[m].freq = f_tuned;

    if ( m < 6 )
    {
      char pLabel[20];
      sprintf(pLabel, "%d - %.2f", (m+1), (double)f_tuned );
      gtk_button_set_label(GTK_BUTTON(button_am_preset[m]), pLabel);    
    }
  }

  /* restore freq tuned before autoscan */
  AM_Tune_Freq(oldFreq);

  /* set GPIO */
  AM_GPIO_Set(oldFreq);
  
  /* update s meter readings */
  AM_SMeter();

  gstr = "AM Mode running .... AutoScan beendet.\0";
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), gstr), gstr);
}

void am_tune_preset_cb(GtkObject *object, gpointer user_data )
{
  int preset = (int)user_data;
  int freq = AM_Preset[ AM_Preset_Bank * 6 + preset ].freq;
  
  waitCTS();
  AM_Tune_Freq(freq);
  Delay (110);
  
  if ( GET_INT_STATUS() )
  {
    AM_Tune_Status(3);
  }

  /* set GPIO */
  AM_GPIO_Set(freq);
  
  /* update s meter readings */
  AM_SMeter();
}

void am_preset_change_bank_cb(GtkObject *object )
{
  int f, m;
  int bank = AM_Preset_Bank + 1;
  int p_start = 6 * bank;
  char pLabel[20];

  if ( (p_start + 6) <= (sizeof(AM_Preset) / sizeof(AM_Preset[0])) )
  {
    AM_Preset_Bank = bank;
  }
  else
  {
    AM_Preset_Bank = 0;
  }

  /* label preset memory bank button */
  sprintf(pLabel, "MEM %d", (AM_Preset_Bank + 1) );
  gtk_button_set_label(GTK_BUTTON(button_am_preset_bank), pLabel);    
 
  for ( m = 0; m < 6; m++ )
  {
    /* get frequency */
    f = AM_Preset[ AM_Preset_Bank * 6 + m ].freq;

    /* relabel preset button */
    if ( f > 0 )
    {
      sprintf(pLabel, "%d - %d kHz", (AM_Preset_Bank * 6 + m+1), f );
      gtk_button_set_label(GTK_BUTTON(button_am_preset[m]), pLabel);    
    }
    else
    {
      sprintf(pLabel, "%d - -------", (AM_Preset_Bank * 6 + m+1));
      gtk_button_set_label(GTK_BUTTON(button_am_preset[m]), pLabel);    
    }
  }  
}

gboolean am_preset_onButtonPressed (GtkObject *object, GdkEventButton *event, gpointer user_data)
{
  int preset = (int)user_data;

  /* single click with the right mouse button? */
  if (event->type == GDK_BUTTON_PRESS && event->button == 3)
  {
    // printf("Single right click on AM preset %d.\n", preset);

    if ( GET_INT_STATUS() )
    {
      AM_Tune_Status(3);
    }

    sleep(1);

    int f_tuned = (R[2] * 256 + R[3]);
    
    AM_Preset[AM_Preset_Bank * 6 + preset].freq = f_tuned;

    if ( preset < 6 )
    {
      char pLabel[20];
      sprintf(pLabel, "%d kHz", f_tuned );
      gtk_button_set_label(GTK_BUTTON(button_am_preset[preset]), pLabel);    
    }

    return TRUE;
  }
  /* single click with the right mouse button? */
  else if (event->type == GDK_BUTTON_PRESS && event->button == 1)
  {
    // printf("Single left click on AM preset %d.\n", preset);

    int freq = AM_Preset[ AM_Preset_Bank * 6 + preset ].freq;
  
    waitCTS();
    AM_Tune_Freq(freq);
    Delay (110);
  
    if ( GET_INT_STATUS() )
    {
      AM_Tune_Status(3);
    }

    AM_SMeter();

    return TRUE;
  }

  return FALSE;
}

void am_freq_band_cb(GtkObject *object, gpointer user_data )
{
  int b = (int)user_data;
  int freq = 0;
  
  /* set prop: band bottom */
  Prop( 0x3400, AM_bands[b].band_bottom);
  Delay(110);
  
  /* set prop: band top */
  Prop( 0x3401, AM_bands[b].band_top);
  Delay(110);

  /* set prop: frequency spacing */
  Prop( 0x3402, AM_bands[b].freq_spacing);
  Delay(110);

  /* set adjustment page increment */
  gtk_adjustment_set_page_increment(GTK_ADJUSTMENT(adj_am_tune),
                                    (double)AM_bands[b].freq_spacing);

  /* tune to band bottom */
  freq = AM_bands[b].band_start_freq;
  AM_Tune_Freq( freq );
  Delay(250);

  if ( GET_INT_STATUS() )
  {
    AM_Tune_Status(3);
  }

  /* set GPIO */
  AM_GPIO_Set(freq);
  
  /* update s meter readings */
  AM_SMeter();
}

void am_smeter_update_cb(GtkObject *object)
{
  AM_SMeter();
}

void change_am_digits( GtkWidget *widget,
                       GtkSpinButton *spin )
{
  int f = startFreqAM; /* init with default AM freq */

  gtk_spin_button_set_digits (GTK_SPIN_BUTTON (entry_am_freq), 0);

  f = gtk_spin_button_get_value_as_int (spin);

  waitCTS();

  AM_Tune_Freq(f);

  Delay (110);

  /* set GPIO */
  AM_GPIO_Set(f);
  
  /* update s meter readings */  AM_SMeter();
}

/* ---------------------------------
    callback routines: general
   --------------------------------- */

void vol_mute_cb(GtkObject *object )
{
  int Arg;
  
  if ( RxVolMuted == 0 ) // not muted -> mute, change button label
  {
    Arg = 3;  
    RxVolMuted = 1;
    gtk_button_set_label(GTK_BUTTON(button_mute), "Un_mute");
  }
  else // muted -> unmute, change label back
  {
    Arg = 0;
    RxVolMuted = 0;
    gtk_button_set_label(GTK_BUTTON(button_mute), "_Mute");
  }

  Prop(0x4001, Arg);
}


void scrollbar_cb_vol(GtkAdjustment *adj, gpointer user_data)
{
  /* change volume */
  waitCTS();
  RX_VOLUME(adj->value);
}

/* ---------------------------------
    callback routines GPIO dev
   --------------------------------- */

void GPIO_CTL(int Arg)
{
  OUTPUT_DEBUG_STRING( "GPIO_CTL()\n" );

  if ( dbgFlag.GPIO )
    printf("GPIO Ctl: %d\n", Arg);

  waitCTS();
  DataOut[1] = 0x80;
  DataOut[2] = Arg;
  I2Cio(2,0);
}

void GPIO_SET(int Arg)
{
  OUTPUT_DEBUG_STRING( "GPIO_SET()\n" );

  if ( dbgFlag.GPIO )
    printf("GPIO Set: %d\n", Arg);

  waitCTS();
  DataOut[1] = 0x81;
  DataOut[2] = Arg;
  I2Cio(2,0);
}

void gpio_pulse(int pulse)
{  
  int cnt;
  
  if ( dbgFlag.GPIO )
    printf("Num of Pulses Arg: %d\n", pulse);

  // start: selecting GPIO1 and set high state
  GPIO_CTL(2);
  GPIO_SET(2);

  for ( cnt = 1; cnt <= pulse; cnt++ )
  {
    // set low
    GPIO_SET(0);
    
    // wait 100 ms
    Delay( GPIOdelay );
    
    // set high again
    GPIO_SET(2); 
  }
}

void AM_GPIO_Set(int freq)
{
  if ( dbgFlag.GPIO )
    printf("AM_GPIO_Set - Freq %d - State: %d\n", freq, GPOstate);

  /* GPO solution with Atmel uC using 1 GPO */
  if ( isGPIOenabled == 1 )
  {
    if ( freq < 522 ) /* LW: 3 pulses */
    {
      if ( GPOstate != 3 )
      {
        gpio_pulse(3);
        GPOstate = 3;
      }
    }
    else if ( (freq >= 522) && (freq < 1711) ) /* MW: 1 pulse */
    {
      if ( GPOstate != 1 )
      {
        gpio_pulse(1);
        GPOstate = 1;
      }
    }
    else if ( freq > 1711 ) /* SW: 2 pulses */
    {
      if ( GPOstate != 2 )
      {
        gpio_pulse(2);
        GPOstate = 2;
      }
    }
  }
  /* GPO solution using 2 GPOs */
  if ( isGPIOenabled == 2 )
  {
    if ( freq < 522 ) /* LW: both low */
    {
      if ( GPOstate != 3 )
      {
        GPIO_CTL(6);
        GPIO_SET(0);
        GPOstate = 3;
      }
    }
    else if ( (freq >= 522) && (freq < 1711) ) /* MW: GPO1 high | GPO2 low */
    {
      if ( GPOstate != 1 )
      {
        GPIO_CTL(6);
        GPIO_SET(2);
        GPOstate = 1;
      }
    }
    else if ( freq > 1711 ) /* SW: both high */
    {
      if ( GPOstate != 2 )
      {
        GPIO_CTL(6);
        GPIO_SET(6);
        GPOstate = 2;
      }
    }
  }
  
  if ( dbgFlag.GPIO )
    printf("AM_GPIO_Set - New State: %d\n", GPOstate);
}

/* callback routines for GPIO evelopment tab */
void gpio_cb(GtkObject *object, gpointer user_data )
{
  int pt;

  /* port */
  pt = (int)user_data;
  
  if ( pt == 1 )
  {
      GPIO_CTL(2);
      GPIO_SET(0);
  }
  else
  {
      GPIO_SET(2);  
  }
}

void gpio_pulse_cb(GtkObject *object, gpointer user_data )
{
  int pulse;

  pulse = (int)user_data;

  gpio_pulse(pulse);
}

/* ---------------------------------
    callback routines general dev
    (taking screenshots, save automated
     rds logs etc.)
   --------------------------------- */

/* Get the current date and format in a printable format */
char* get_date()
{
  struct tm* t;
  time_t now;
  time( &now );
  t = localtime( &now );

  char* date = malloc(32*sizeof(char));
  sprintf(date, "%d-%d-%d", t->tm_mday,t->tm_mon+1,t->tm_year+1900);
  return date;
}

char* get_datetime()
{
  struct tm* t;
  time_t now;
  time( &now );
  t = localtime( &now );

  char* date = malloc(64*sizeof(char));
  sprintf(date, "%d-%d-%d_%d:%d:%d", t->tm_mday,t->tm_mon+1,t->tm_year+1900,t->tm_hour,t->tm_min,t->tm_sec);
  return date;
}

GdkPixbuf *get_screenshot(GtkObject *wobject)
{
    GdkPixbuf *screenshot;
    GdkWindow *root_window;
    GdkWindow *active_window;
   GdkRectangle rect;
    
    gint x_orig, y_orig;
    gint width, height;

    /* get root windows (desktop) */
    root_window = gdk_get_default_root_window();

    /* get active window, i.e. si4735radio application */
    active_window = GTK_WIDGET(wobject)->window;

    /* get size of active windows */
    gdk_window_get_frame_extents(active_window, &rect);
             
    /* get infos about root window */
    gdk_drawable_get_size (root_window, &width, &height);      
    gdk_window_get_origin (root_window, &x_orig, &y_orig);

//    screenshot = gdk_pixbuf_get_from_drawable (NULL, root_window, NULL,
//                                           x_orig, y_orig, 0, 0, width, height);

    screenshot = gdk_pixbuf_get_from_drawable (NULL, root_window, NULL,
                                           rect.x, rect.y, 0, 0, rect.width, rect.height);
    return screenshot;
}

void appdbgFlag_take_screenshot(GtkObject *wobject)
{
    GdkPixbuf *screenshot;
    
    Delay(25);
    
    gtk_widget_set_sensitive(GTK_WIDGET(tb_screenshot), FALSE);

    while ( gtk_events_pending() ) 
    {
      gtk_main_iteration();
    }
    
    screenshot = get_screenshot(wobject);

    gtk_widget_set_sensitive(GTK_WIDGET(tb_screenshot), TRUE);

//    gdk_pixbuf_save(screenshot, "/tmp/si4735-001.jpg", "jpeg", NULL, "quality", "70", NULL);
    gdk_pixbuf_save(screenshot, "/tmp/si4735-001.png", "png", NULL, NULL);
}

void appdbgFlag_take_screenshot_cb(GtkObject *wobject, gpointer user_data)
{
    appdbgFlag_take_screenshot(wobject);
}

/* ============================
      .::. DIALOG BOXES .::. 
   ============================ */

void show_about_dialog(GtkWidget *widget, gpointer data)
{
  const gchar *authors[] =
  {
    "Tobias Stoeber <tobi@to-st.de>\n",
    _("Si4735 FM/AM Radio Linux software is derived from Delphi software by:\n"
    "  * Burkhard Kanka <B.Kainka@t-online.de>\n"
    "    http://www.elexs.de/SI4735c.html\n"),
    _("Contributions and testing:\n"
    "  * Ralf Beesner\n"),
    NULL
  };

  gchar *comments = _("Si4735 PC Radio with USB V2 Software\n(based on libftdi or libftd2xx)"); 

  /* TRANSLATORS: Replace this string with your names, one name per line. */
  gchar *translators =  _( "translator-credits" );

  /* read logo from PNG file */
  // GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file("si4735-32x32.png", NULL);

  /* get logo data from included xpm string */
  GdkPixbuf *pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)bmp_si4735);

  GtkWidget *about_dialog = gtk_about_dialog_new();

  gtk_container_set_border_width ( ( GtkContainer*)about_dialog , 20 );

  gtk_about_dialog_set_program_name(		GTK_ABOUT_DIALOG(about_dialog), PROGRAM);
  gtk_about_dialog_set_version(			GTK_ABOUT_DIALOG(about_dialog), VERSION); 
  gtk_about_dialog_set_copyright(		GTK_ABOUT_DIALOG(about_dialog), "Copyright (C) 2011,2012 Tobias Stoeber et. al.");
  gtk_about_dialog_set_comments(		GTK_ABOUT_DIALOG(about_dialog), comments);
  gtk_about_dialog_set_website(			GTK_ABOUT_DIALOG(about_dialog), "http://www.to-st.de/content/projects/si4735lin/");
  gtk_about_dialog_set_license ( 		GTK_ABOUT_DIALOG(about_dialog), "Si4735 FM/AM Radio\n\nCopyright (C) 2011,2012 Tobias Stoeber\n\nThis program is free software; you can redistribute it and/or\nmodify it under the terms of the GNU General Public License\nas published by the Free Software Foundation; version 2\nof the License.\n\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\nYou should have received a copy of the GNU General Public License\nalong with this program; if not, write to the Free Software\nFoundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA." );
  gtk_about_dialog_set_authors ( 		GTK_ABOUT_DIALOG(about_dialog), authors );
  gtk_about_dialog_set_translator_credits ( 	GTK_ABOUT_DIALOG(about_dialog), translators );
  gtk_about_dialog_set_logo(			GTK_ABOUT_DIALOG(about_dialog), pixbuf);
  g_object_unref(pixbuf), pixbuf = NULL;

  gtk_dialog_run(GTK_DIALOG (about_dialog));

  gtk_widget_destroy(about_dialog);
}

void show_spinner_dialog(void)
{
    GtkWidget *dlg;
    GtkWidget *c_area;
    GtkWidget *vbox;
    GtkWidget *spinner;
    GtkWidget *tmp_lbl;

    vbox = gtk_vbox_new(FALSE, 0);

    gchar tmp_str[30];

    /* dialog window */
    dlg = gtk_dialog_new_with_buttons(_("Spinner"), NULL, GTK_DIALOG_MODAL, GTK_STOCK_STOP, GTK_RESPONSE_REJECT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, NULL);

    gtk_window_set_default_size(GTK_WINDOW(dlg), 150, 150);		// default size
    gtk_window_set_position(GTK_WINDOW(dlg), GTK_WIN_POS_CENTER);	// position
    gtk_container_set_border_width(GTK_CONTAINER(dlg), 3);		// border
    gtk_window_set_icon(GTK_WINDOW(dlg), gdk_pixbuf_new_from_xpm_data((const char **)bmp_si4735));

    /* get content area */
    c_area = gtk_dialog_get_content_area(GTK_DIALOG (dlg));

    spinner = gtk_spinner_new();    
    gtk_spinner_start (GTK_SPINNER(spinner));

    sprintf(tmp_str, "<b>Radio IC Information</b>");
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_markup( GTK_LABEL(tmp_lbl), tmp_str); 
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),0.5,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_box_pack_start_defaults(GTK_BOX(vbox), tmp_lbl);    

    gtk_box_pack_start_defaults(GTK_BOX (vbox), spinner);

    gtk_box_pack_start_defaults(GTK_BOX (c_area), vbox);

    gtk_widget_show_all(vbox);
    gtk_widget_show_all(dlg);

    si4735radio_presets_save();

//    if (gtk_dialog_run(GTK_DIALOG (dlg)) == GTK_RESPONSE_ACCEPT)
//	printf("jo\n");

//    gtk_widget_destroy(dlg);

  g_signal_connect_swapped (dlg, "response",
        	                  G_CALLBACK (gtk_widget_destroy),
                	          dlg);
}

GtkWidget *am_prefs_spin_button[32];
GtkWidget *fm_prefs_spin_button[32];

void apply_prefs()
{
  int i, j, val = 0;
  int am_arr_size = sizeof( SI4735_prop_am )/sizeof(  SI4735_prop_am[0]);
  int fm_arr_size = sizeof( SI4735_prop_fm )/sizeof(  SI4735_prop_fm[0]);

  /* AM properties */
  if ( isModeAM )
  {  
    for ( i = 0; i < am_arr_size; i++ )
    {
      if ( SI4735_prop_am[i].isDiscrete )
      {
        char *text =  gtk_combo_box_get_active_text(GTK_COMBO_BOX(am_prefs_spin_button[i]));

        for ( j = 0; j < 10; j++ )
        {
          if ( SI4735_prop_am[i].val_disc[j] != NULL )
          {
            if ( strcmp( (char *)SI4735_prop_am[i].val_disc[j], (char *)text) == 0 )
            {
              val = j;
            }
          }
        }
      }
      else
      {
        val = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(am_prefs_spin_button[i]));
      }
      
      // printf("AM Prop: %x (%s) = %d\n", SI4735_prop_am[i].prop, SI4735_prop_am[i].name, val);

      /* keep current value in dataset */
      SI4735_prop_am[i].val_curr = val;      

      /* set property */
      Prop(SI4735_prop_am[i].prop, val);
    }
  }

  /* FM properties */
  if ( isModeFM )
  {  
    for ( i = 0; i < fm_arr_size; i++ )
    {
      if ( SI4735_prop_fm[i].isDiscrete )
      {
        char *text =  gtk_combo_box_get_active_text(GTK_COMBO_BOX(fm_prefs_spin_button[i]));

        for ( j = 0; j < 10; j++ )
        {
          if ( SI4735_prop_fm[i].val_disc[j] != NULL )
          {
            if ( strcmp( (char *)SI4735_prop_fm[i].val_disc[j], (char *)text) == 0 )
            {
              val = j;
            }
          }
        }
      }
      else
      {
        val = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(fm_prefs_spin_button[i]));
      }
      
      // printf("FM Prop: %x (%s) = %d\n", SI4735_prop_fm[i].prop, SI4735_prop_fm[i].name, val);

      /* keep current value in dataset */
      SI4735_prop_fm[i].val_curr = val;      

      /* set property */
      Prop(SI4735_prop_fm[i].prop, val);
    }
  }
}

void show_preferences(void)
{
    GtkWidget *dlg;
    GtkWidget *c_area;
    GtkWidget *notebook;
    GtkWidget *info_box, *fm_prefs_box, *am_prefs_box, *misc_prefs_box;
    GtkWidget *tmp_lbl;
    GtkWidget *tmp_lbl_unit;
    GtkWidget *table;

    int i, val;
    
    /* dialog window */
    dlg = gtk_dialog_new_with_buttons(_("Preferences"), NULL, 0, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_APPLY, GTK_RESPONSE_APPLY, NULL);

    gtk_window_set_default_size(GTK_WINDOW(dlg), 450, 250);		// default size
    gtk_window_set_position(GTK_WINDOW(dlg), GTK_WIN_POS_CENTER);	// position
    gtk_container_set_border_width(GTK_CONTAINER(dlg), 3);		// border
    gtk_window_set_icon(GTK_WINDOW(dlg), gdk_pixbuf_new_from_xpm_data((const char **)bmp_si4735));

    /* new tooltips */
    prefs_tips = gtk_tooltips_new();

    /* get content area */
    c_area = gtk_dialog_get_content_area(GTK_DIALOG (dlg));

    info_box = gtk_vbox_new(FALSE, 0);
    fm_prefs_box = gtk_vbox_new(FALSE, 0);
    am_prefs_box = gtk_vbox_new(FALSE, 0);
    misc_prefs_box = gtk_vbox_new(FALSE, 0);

    /* new notebook widget */
    notebook = gtk_notebook_new();
    gtk_notebook_set_tab_pos( GTK_NOTEBOOK(notebook), GTK_POS_TOP);
    gtk_box_pack_start_defaults(GTK_BOX (c_area), notebook);

    /* INFO box */
    tmp_lbl = gtk_label_new(_("Info"));
    // gtk_label_set_angle( GTK_LABEL(tmp_lbl), 90);
    gtk_notebook_append_page(GTK_NOTEBOOK (notebook), info_box, tmp_lbl );

    /* FM prefs box */
    tmp_lbl = gtk_label_new(_("FM Properties"));
    // gtk_label_set_angle( GTK_LABEL(tmp_lbl), 90);
    gtk_notebook_append_page(GTK_NOTEBOOK (notebook), fm_prefs_box, tmp_lbl );

    /* AM prefs box */
    tmp_lbl = gtk_label_new(_("AM Properties"));
    // gtk_label_set_angle( GTK_LABEL(tmp_lbl), 90);
    gtk_notebook_append_page(GTK_NOTEBOOK (notebook), am_prefs_box, tmp_lbl );

    /* ------------------
       .::. Info Box .::.
       ------------------ */

    /* get ic info */
    if ( isModeFM || isModeAM )
    { 
      GetRev();

      sprintf(info_si.ic, "47%d", R[1]);
      sprintf(info_si.fw, "%c%c", R[2], R[3]);
      if ( isModeFM )
      {  
        sprintf(info_si.comp_fm, "%c%c", R[6], R[7]);
      }
      else
      {
        sprintf(info_si.comp_am, "%c%c", R[6], R[7]);      
      }
      sprintf(info_si.chip_rev, "%c", R[8]);
    }
    
    /* do something useful */
    gchar tmp_str[30]; 

    table = gtk_table_new(5, 2, FALSE);

    /* headline */
    sprintf(tmp_str, "<b>Radio IC Information</b>");
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_markup( GTK_LABEL(tmp_lbl), tmp_str); 
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),0.5,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_box_pack_start_defaults(GTK_BOX(info_box), tmp_lbl);    

    /* info: ic */
    sprintf(tmp_str, "Radio IC:");
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str); 
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),1.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 0, 1, 0, 1);
    sprintf(tmp_str, "%s", (char *) info_si.ic);
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str); 
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),0.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 1, 2, 0, 1);

    /* info: firmware */
    sprintf(tmp_str, "Firmware:");
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str);
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),1.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 0, 1, 1, 2);
    sprintf(tmp_str, "%s", (char *) info_si.fw);
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str);
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),0.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 1, 2, 1, 2);

    /* info: component FM RX */
    sprintf(tmp_str, "FM RX Component:");
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str);
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),1.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 0, 1, 2, 3);
    sprintf(tmp_str, "%s", (char *) info_si.comp_fm);
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str);
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),0.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 1, 2, 2, 3);

    /* info: component AM RX */
    sprintf(tmp_str, "LW/AM/SW RX Component:");
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str);
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),1.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 0, 1, 3, 4);
    sprintf(tmp_str, "%s", (char *) info_si.comp_am);
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str);
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),0.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 1, 2, 3, 4);

    /* info: chip rev */
    sprintf(tmp_str, "Chip Rev:");
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str);
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),1.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 0, 1, 4, 5);
    sprintf(tmp_str, "%s", (char *) info_si.chip_rev);
    tmp_lbl = gtk_label_new(NULL);
    gtk_label_set_text( GTK_LABEL(tmp_lbl), tmp_str);
    gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(tmp_lbl),0.0,0.5);
    gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
    gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 1, 2, 4, 5);

    gtk_box_pack_start(GTK_BOX(info_box), table, FALSE, FALSE, 0);    

    /* ------------------
       .::. FM Prefs .::.
       ------------------ */

    int fm_arr_size = sizeof( SI4735_prop_fm )/sizeof(  SI4735_prop_fm[0]);
    int j;

    table = gtk_table_new(fm_arr_size, 4, FALSE);

    for ( i = 0; i < fm_arr_size; i++ )
    {
      tmp_lbl = gtk_label_new(_(SI4735_prop_fm[i].name2));
      tmp_lbl_unit = gtk_label_new(_(SI4735_prop_fm[i].units));
    
      if ( SI4735_prop_fm[i].isDiscrete )
      {
        fm_prefs_spin_button[i] = gtk_combo_box_new_text();

        for ( j = 0; j < 10; j++ )
        {
          if ( SI4735_prop_fm[i].val_disc[j] != NULL )
          {
            gtk_combo_box_append_text(GTK_COMBO_BOX(fm_prefs_spin_button[i]), (char *)SI4735_prop_fm[i].val_disc[j]);
          }
        }
      }
      else
      {
        fm_prefs_spin_button[i] = gtk_spin_button_new_with_range(SI4735_prop_fm[i].val_min, SI4735_prop_fm[i].val_max, SI4735_prop_fm[i].val_step);
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(fm_prefs_spin_button[i]), SI4735_prop_fm[i].val_def);
      }
      
      if ( isModeFM )
      {
        val = Get_Prop(SI4735_prop_fm[i].prop);
        if ( SI4735_prop_fm[i].isDiscrete )
        {
          gtk_combo_box_set_active(GTK_COMBO_BOX(fm_prefs_spin_button[i]), val);
        }
        else
        {
          gtk_spin_button_set_value(GTK_SPIN_BUTTON(fm_prefs_spin_button[i]), val);
        }
      }
      
      /* label: prop name */
      gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_RIGHT);
      gtk_misc_set_alignment(GTK_MISC(tmp_lbl),1.0,0.5);
      gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
      gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 0, 1, i, i+1);


      /* label: spin button */
      // gtk_misc_set_alignment(GTK_MISC(am_prefs_spin_button[i]),0.0,0.5);
      // gtk_misc_set_padding(GTK_MISC(am_prefs_spin_button[i]),5,5);

      gtk_tooltips_set_tip (GTK_TOOLTIPS (prefs_tips), fm_prefs_spin_button[i],
                           SI4735_prop_fm[i].desc, "");

      gtk_table_attach_defaults(GTK_TABLE(table), fm_prefs_spin_button[i], 1, 2, i, i+1);

      /* label: unit */
      gtk_label_set_justify( GTK_LABEL(tmp_lbl_unit), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment(GTK_MISC(tmp_lbl_unit),0.0,0.5);
      gtk_misc_set_padding(GTK_MISC(tmp_lbl_unit),5,5);
      gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl_unit, 2, 3, i, i+1);
    }

    // gtk_box_pack_start_defaults(GTK_BOX (fm_prefs_box), table);
    if ( isModeFM )
    {
      gtk_box_pack_start(GTK_BOX (fm_prefs_box), table, FALSE, FALSE, 0);
    }
    else
    {
      gtk_box_pack_start_defaults(GTK_BOX (fm_prefs_box), gtk_label_new(_("Only available in FM Mode.")));    
    }
    /* ------------------
       .::. AM Prefs .::.
       ------------------ */

    int am_arr_size = sizeof( SI4735_prop_am )/sizeof(  SI4735_prop_am[0]);

    table = gtk_table_new(am_arr_size, 4, FALSE);

    for ( i = 0; i < am_arr_size; i++ )
    {
      tmp_lbl = gtk_label_new(_(SI4735_prop_am[i].name2));
      tmp_lbl_unit = gtk_label_new(_(SI4735_prop_am[i].units));
    
      if ( SI4735_prop_am[i].isDiscrete )
      {
        am_prefs_spin_button[i] = gtk_combo_box_new_text();

        for ( j = 0; j < 10; j++ )
        {
          if ( SI4735_prop_am[i].val_disc[j] != NULL )
          {
            gtk_combo_box_append_text(GTK_COMBO_BOX(am_prefs_spin_button[i]), (char *)SI4735_prop_am[i].val_disc[j]);
          }
        }
      }
      else
      {
        am_prefs_spin_button[i] = gtk_spin_button_new_with_range(SI4735_prop_am[i].val_min, SI4735_prop_am[i].val_max, SI4735_prop_am[i].val_step);
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(am_prefs_spin_button[i]), SI4735_prop_am[i].val_def);
      }
      
      if ( isModeAM )
      {
        val = Get_Prop(SI4735_prop_am[i].prop);
        if ( SI4735_prop_am[i].isDiscrete )
        {
          gtk_combo_box_set_active(GTK_COMBO_BOX(am_prefs_spin_button[i]), val);
        }
        else
        {
          gtk_spin_button_set_value(GTK_SPIN_BUTTON(am_prefs_spin_button[i]), val);
        }
      }

      /* label: prop name */
      gtk_label_set_justify( GTK_LABEL(tmp_lbl), GTK_JUSTIFY_RIGHT);
      gtk_misc_set_alignment(GTK_MISC(tmp_lbl),1.0,0.5);
      gtk_misc_set_padding(GTK_MISC(tmp_lbl),5,5);
      gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl, 0, 1, i, i+1);

      /* label: spin button */
      // gtk_misc_set_alignment(GTK_MISC(am_prefs_spin_button[i]),0.0,0.5);
      // gtk_misc_set_padding(GTK_MISC(am_prefs_spin_button[i]),5,5);
      gtk_tooltips_set_tip (GTK_TOOLTIPS (prefs_tips), am_prefs_spin_button[i],
                           SI4735_prop_am[i].desc, "");

      gtk_table_attach_defaults(GTK_TABLE(table), am_prefs_spin_button[i], 1, 2, i, i+1);

      /* label: unit */
      gtk_label_set_justify( GTK_LABEL(tmp_lbl_unit), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment(GTK_MISC(tmp_lbl_unit),0.0,0.5);
      gtk_misc_set_padding(GTK_MISC(tmp_lbl_unit),5,5);
      gtk_table_attach_defaults(GTK_TABLE(table), tmp_lbl_unit, 2, 3, i, i+1);
    }

    if ( isModeAM )
    {
      gtk_box_pack_start_defaults(GTK_BOX (am_prefs_box), table);
    }
    else
    {
      gtk_box_pack_start_defaults(GTK_BOX (am_prefs_box), gtk_label_new(_("Only available in AM/SW/LW Mode."))); 
    }
    
    gtk_widget_show_all(notebook);

    int resp;

    /* show ("run") the dialog window, apply prefs on ACCEPT (close) or APPLY (keeping dialog open) */
    while ( (resp = gtk_dialog_run(GTK_DIALOG (dlg))) )
    {
      if ( resp == GTK_RESPONSE_ACCEPT )
      {
        apply_prefs();
        break;
      }
      else if ( resp == GTK_RESPONSE_APPLY )
      {
        apply_prefs();
      }
      else if ( resp == GTK_RESPONSE_REJECT )
      {
        break;
      }
    }
    
    gtk_widget_destroy(dlg);
}

void show_cli_help(void)
{
  printf("\nUsage: si4735 [OPTION]\n");
  printf("\nthe following command line options are available: \n\n");
  printf("\t-h         show short help (this text)\n");
  printf("\t-l         enable RDS group logging\n");
  printf("\t-r  file   filename for RDS logfile\n");
  printf("\t-D  arg    debug options (arg = GPO1, GPO2, RT, TMC)\n");  
  printf("\t-F  arg    format for RDS log, 'hex' (default) or 'raw'\n");
  printf("\t-G         enable GPIO (0 = disabled, 1 = 1 GPO+Atmel, 2 = 2 GPOs)\n");
  printf("\t-L         list FTDI devices and exit\n");
  printf("\t-N  arg    USB serial of FTDI device\n");
  printf("\t-R         enable RDS logging\n");
  printf("\t-S         enable screen capture / screenshot feature\n");
  printf("\t-v         show version info\n\n");

  exit(EXIT_SUCCESS);
}

/* ============================
      .::. MAIN PROGRAM .::. 
   ============================ */

int main(int argc, char **argv)
{

  /* print short info line about program */
  printf("%s\n\n", INFO);

  /* ------------------------------- */
  /* command line options processing */
  /* ------------------------------- */  
  int c;
  opterr = 0;
     
  while ((c = getopt (argc, argv, "hlr:D:F:G:LN:RS")) != -1)
    switch (c)
    {
      case 'h':
        show_cli_help();
        break;
      case 'l':
        startRDSlog = 1;
        break;
      case 'R':
        startRDS = 1;
        break;
      case 'r':
        fileRDSlog = optarg;
        printf("RDS logfile filename is %s\n", fileRDSlog);
        break;
      case 'D':
        if ( strcmp(optarg, "RT") > - 1)
        {
          dbgFlag.RTplus = 1;
          printf("Debug - RDS: debugging RT plus enabled.\n");
        }
        else if ( strcmp(optarg, "GPO1") > - 1)
        {
          dbgFlag.GPIO = 1;
          printf("Debug - GPIO: debugging GPIO 1 solution enabled.\n");
        }
        else if ( strcmp(optarg, "GPO2") > - 1)
        {
          dbgFlag.GPIO = 2;
          printf("Debug - GPIO: debugging GPIO 1 + 2 solution enabled.\n");
        }
        else if ( strcmp(optarg, "TMC") > - 1)
        {
          dbgFlag.TMC = 1;
          printf("Debug TMC: debugging TMC groups enabled.\n");
        }
        break;
      case 'F':
        rdsLogFormat = optarg;
        printf("RDS logfile format is %s\n", rdsLogFormat);
        break;
      case 'G':
        isGPIOenabled = (int) atol(optarg);
        printf("GPIO enabled mode %d.\nBE CAREFUL!!!\n", isGPIOenabled);
        break;
      case 'L':
        printf("Listing FT-Devices ....\n");
        FTDI_Find_All(-1);
        break;
      case 'N':
        ser = (char *)optarg;
        use_iSerial = 1;
        printf("FTDI device serial %s\n", ser);
        break;
      case 'S':
        dbgFlag.ScreenCap = 1;
        break;
      case '?':
        if ( (optopt == 'F') || (optopt == 'r') )
          fprintf (stderr, "Option -%c requires an argument.\n", optopt);
        else if (isprint (optopt))
          fprintf (stderr, "Unknown option `-%c'.\n", optopt);
        else
          fprintf (stderr,
                    "Unknown option character `\\x%x'.\n",
                    optopt);
        return 1;
      default:
        abort();
    }

  /* load presets */
  si4735radio_presets_load();  
 
  /* load setting */
  si4735radio_settings_load();

  /* ---------------
     FTDI Init etc.
     --------------- */   
  FTDI_Init();
  FTDI_Connect();

  /* ---------------------
     init some vars 
     --------------------- */
  rdsaktiv = 0;
  Zeit = 0;
  Minuten = 0;
  Stunden = 0;
  RDS_PI = 0;
  TP = 0; TA = 0; 
  MS = 0; PTY = 0;
  FT_port = 21;        /* 0 = TXD, Reset; 2 = DTR = SDA, 2 = RTS = SCL */

  /* pre-fill DataOut[] */
  int l;
  for ( l = 0; l <= 13; l++ )
  {
    DataOut[l] = 0x0;
  }

  /* pre-fill Codearray / CodearrayPS */
  for ( l = 0; l <= 31; l++ )
  {
    Codearray_RT_A[l] = 0x20;
    Codearray_RT_B[l] = 0x20;
    CodearrayRT2[l] = 0x20;
    CodearrayPS[l] = 0x20;
    tmpCodearrayPS[l] = 0x20;
  }

  for ( l = 32; l <= 63; l++ )
  {
    Codearray_RT_A[l] = 0x20;
    Codearray_RT_B[l] = 0x20;
    CodearrayRT2[l] = 0x20;
  }

  
  /* rest chipset */
  ChipReset();

  /* -----------
      GTK+ part
     ----------- */

  /* initialise Gtk */
  gtk_init(&argc, &argv);

  /* create new top level window */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  /* create global VBox, contains notebook and elements for all tabs (e.g. volume slider) */   
  global_vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_set_border_width (GTK_CONTAINER (global_vbox), 0);
  gtk_container_add (GTK_CONTAINER (window), global_vbox);

  tb_vbox = gtk_vbox_new(FALSE, 5);

  /*
   *  Toolbar 
   */

  toolbar = gtk_toolbar_new();
  // gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_ICONS);

  gtk_container_set_border_width(GTK_CONTAINER(toolbar), 0);

  gtk_toolbar_set_orientation (GTK_TOOLBAR (toolbar), GTK_ORIENTATION_HORIZONTAL);
  gtk_toolbar_set_style (GTK_TOOLBAR (toolbar), GTK_TOOLBAR_ICONS); /* ICONS, TEXT, BOTH */
  gtk_container_set_border_width (GTK_CONTAINER (toolbar), 5);

/*
  tb_new = gtk_tool_button_new_from_stock(GTK_STOCK_NEW);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_new, -1);
*/

/*
  tb_open = gtk_tool_button_new_from_stock(GTK_STOCK_OPEN);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_open, -1);
*/

/*
  tb_save = gtk_tool_button_new_from_stock(GTK_STOCK_SAVE);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_save, -1);
*/

  tb_sep = gtk_separator_tool_item_new();
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_sep, -1); 

  tb_prefs = gtk_tool_button_new_from_stock(GTK_STOCK_PREFERENCES);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_prefs, -1);

  tb_sep = gtk_separator_tool_item_new();
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_sep, -1); 

  GtkWidget *icon_new;
  icon_new = gtk_image_new_from_file("icon_screenshot-32x32.png"); // /usr/share/icons/gnome/48x48/apps/applets-screenshooter.png");
  tb_screenshot = gtk_tool_button_new(icon_new, NULL);

  if ( dbgFlag.ScreenCap == 1 )
    gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_screenshot, -1);

  if ( dbgFlag.ScreenCap )
  {
    tb_sep = gtk_separator_tool_item_new();
    gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_sep, -1); 
  }
  
  tb_help = gtk_tool_button_new_from_stock(GTK_STOCK_HELP);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_help, -1);
  gtk_widget_set_sensitive(GTK_WIDGET(tb_help), FALSE);

  tb_info = gtk_tool_button_new_from_stock(GTK_STOCK_INFO);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_info, -1);
  gtk_widget_set_sensitive(GTK_WIDGET(tb_info), FALSE);

  tb_about = gtk_tool_button_new_from_stock(GTK_STOCK_ABOUT);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_about, -1);

  tb_sep = gtk_separator_tool_item_new();
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_sep, -1); 

  tb_exit = gtk_tool_button_new_from_stock(GTK_STOCK_QUIT);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_exit, -1);

  gtk_box_pack_start(GTK_BOX(global_vbox), toolbar, FALSE, FALSE, 5);

  /* conenct ToolBar signals */
  g_signal_connect(G_OBJECT(tb_exit), "clicked", 
        G_CALLBACK(gtk_main_quit), NULL);

  g_signal_connect(G_OBJECT(tb_about), "clicked", 
        GTK_SIGNAL_FUNC(show_about_dialog), NULL);

  g_signal_connect(G_OBJECT(tb_prefs), "clicked", 
        GTK_SIGNAL_FUNC(show_preferences), NULL);

  gtk_signal_connect_object(GTK_OBJECT(tb_screenshot), "clicked",
                             GTK_SIGNAL_FUNC(appdbgFlag_take_screenshot_cb), GTK_OBJECT(window));

  g_signal_connect_swapped(G_OBJECT(window), "destroy",
        G_CALLBACK(gtk_main_quit), NULL);
  /* END: toolbar */

  /* Create a new notebook */
  notebook = gtk_notebook_new ();
  gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_LEFT);

  /* pack notebook frames into global vbox */
  gtk_box_pack_start(GTK_BOX(global_vbox), notebook, FALSE, FALSE, 0);
  gtk_widget_show(notebook);

  /* create main VBox: FM radio */
  main_vbox_fm = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER (main_vbox_fm), 5);

  /* create main VBox: FM radio - RDS detail info */
  main_vbox_rds = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER (main_vbox_rds), 5);

  /* create main VBox: AM radio */
  main_vbox_am = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER (main_vbox_am), 5);

  /* create main VBox: GPIO develeopment */
  main_vbox_gp = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER (main_vbox_gp), 5);

  /* create labels for notebook tabs */
  notebook_label_fm = gtk_label_new("FM Radio");
  gtk_label_set_angle( GTK_LABEL(notebook_label_fm), 90);
  notebook_label_rds = gtk_label_new("FM + RDS Info");
  gtk_label_set_angle( GTK_LABEL(notebook_label_rds), 90);
  notebook_label_am = gtk_label_new("AM Radio");
  gtk_label_set_angle( GTK_LABEL(notebook_label_am), 90);
  notebook_label_gp = gtk_label_new("GPIO");
  gtk_label_set_angle( GTK_LABEL(notebook_label_gp), 90);

  /* add vboxes into notebook pages */
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), main_vbox_fm, notebook_label_fm);
  // gtk_notebook_append_page (GTK_NOTEBOOK (notebook), main_vbox_rds, notebook_label_rds);
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), main_vbox_am, notebook_label_am);
  if ( (isGPIOenabled != 0) && (dbgFlag.GPIO != 0) )
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook), main_vbox_gp, notebook_label_gp);

  /* create tooltips */
  button_tips = gtk_tooltips_new();
  entry_tips = gtk_tooltips_new();

  /* -------------------------------
      GUI frames - general part 
     ------------------------------- */
     
  /* create frame for volume setting */
  frame_vol = gtk_frame_new("Lautstärke");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_vol), GTK_SHADOW_ETCHED_IN);

  /* -------------------------------
      GUI frames - FM part
     ------------------------------- */

  /* create frame for tuning / freq setting */
  frame_fm_tune = gtk_frame_new("Frequenz / Tuning");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_fm_tune), GTK_SHADOW_ETCHED_IN);

  /* create frame for RDS info etc. */
  frame_fm_rds = gtk_frame_new("RDS Basis Info");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_fm_rds), GTK_SHADOW_ETCHED_IN);

  /* create frame for S-meter info  */
  frame_fm_smeter = gtk_frame_new("FM Radio Init / S-Meter");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_fm_smeter), GTK_SHADOW_ETCHED_IN);

  /* -------------------------------
      GUI frames - AM part
     ------------------------------- */

  /* create frame for tuning / freq setting */
  frame_am_tune = gtk_frame_new("Frequenz / Tuning");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_am_tune), GTK_SHADOW_ETCHED_IN);

  /* create frame for AM Presets */
  frame_am_preset = gtk_frame_new("Senderspeicher");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_am_preset), GTK_SHADOW_ETCHED_IN);

  /* create frame for AM band selection */
  frame_am_band = gtk_frame_new("Band");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_am_band), GTK_SHADOW_ETCHED_IN);

  /* create frame for S-meter info  */
  frame_am_smeter = gtk_frame_new("AM Radio Init / S-Meter");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_am_smeter), GTK_SHADOW_ETCHED_IN);

  /* -------------------------------
      buttons - FM part
     ------------------------------- */

  button_fm_start  = gtk_button_new_with_label("FM Radio starten");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_start,
                           "Startet Si4735 PC Radio im FM Modus / Initialisierung", "");

  button_fm_seek_down  = gtk_button_new_with_label("- Suchlauf");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_seek_down,
                           "Suchlauf nach unten", "");

  button_fm_seek_up  = gtk_button_new_with_label("+ Suchlauf");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_seek_up,
                           "Suchlauf nach oben", "");

  button_fm_autoscan  = gtk_button_new_with_label("AutoScan");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_autoscan,
                           "Suchlauf uber das gesamte FM Band, belegt Speicherplätze", "");

  button_fm_band_norm  = gtk_button_new_with_label("FM INT");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_band_norm,
                           "Setzt Grenzen des FM Bereiches auf 87.5 - 108.0 MHz (EU/US)", "");

  button_fm_band_jp  = gtk_button_new_with_label("FM JP");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_band_jp,
                           "Setzt Grenzen des FM Bereiches auf 76 - 90 MHz (Japan)", "");
                           
  button_fm_band_oirt  = gtk_button_new_with_label("FM OIRT");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_band_oirt,
                           "Setzt Grenzen des FM Bereiches auf 65.8 - 74.0 MHz (Osteuropa)", "");
                           
  button_rds  = gtk_button_new_with_label("RDS starten");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_rds,
                           "Startet bzw. stoppt den RDS Decoder / die RDS Anzeige", "");

  button_fm_smeter_update  = gtk_button_new_with_label("S-Meter aktualisieren");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_smeter_update,
                           "Aktualisiert die S-Meter Anzeige (RSSI und SNR)", "");
  /* -------------------------------
      buttons - AM part
     ------------------------------- */

  button_am_start  = gtk_button_new_with_label("AM Radio starten");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_start,
                           "Startet Si4735 PC Radio im AM Modus / Initialisierung", "");

  button_am_seek_down  = gtk_button_new_with_label("- Suchlauf");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_seek_down,
                           "Suchlauf nach unten", "");

  button_am_seek_up  = gtk_button_new_with_label("+ Suchlauf");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_seek_up,
                           "Suchlauf nach oben", "");

  button_am_autoscan  = gtk_button_new_with_label("AutoScan");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_autoscan,
                           "Suchlauf uber das gesamte FM Band, belegt Speicherplätze", "");

  button_am_smeter_update  = gtk_button_new_with_label("S-Meter akt.");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_smeter_update,
                           "Aktualisiert die S-Meter Anzeige (RSSI, SNR und Cap)", "");

  button_am_tune_1up  = gtk_button_new_with_label("+1 kHz");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_tune_1up,
                           "tune + 1kHz", "");

  button_am_tune_1down  = gtk_button_new_with_label("-1 kHz");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_tune_1down,
                           "tune - 1kHz", "");

  button_am_tune_5up  = gtk_button_new_with_label("+5 kHz");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_tune_5up,
                           "tune + 1kHz", "");

  button_am_tune_5down  = gtk_button_new_with_label("-5 kHz");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_tune_5down,
                           "tune - 1kHz", "");

  button_am_tune_9up  = gtk_button_new_with_label("+9 kHz");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_tune_9up,
                           "tune + 9kHz", "");

  button_am_tune_9down  = gtk_button_new_with_label("-9 kHz");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_tune_9down,
                           "tune - 9kHz", "");

  /* -------------------------------
      buttons - general part
     ------------------------------- */

  button_mute  = gtk_toggle_button_new_with_mnemonic("_Mute");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_mute,
                           "Stummschaltung (Mute), Tastenkombi Alt+M", "");

  button_quit  = gtk_button_new_with_label("Radio verlassen");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_quit,
                           "Beendet die Radiosoftware, Radio spielt weiter", "");

  button_exit  = gtk_button_new_with_label("Radio beenden");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_exit,
                           "Beendet die Radiosoftware, Radio wird zurückgesetzt", "");

  /* -------------------------------
      Adjustments
     ------------------------------- */
                           
  /* volume adjustment, RX_VOLUME: 0 <= VOL <= 63, default: startVolume */
  adj_vol = gtk_adjustment_new(startVolume, 0.0, 64.0, 1.0, 2.0, 1.0);

  /* frequency adjustment, FM (87.50 - 108.00), default startFreqFM */
  adj_fm_tune = gtk_adjustment_new( (float)(startFreqFM/100), 87.5, 108.0, 0.1, 0.5, 0.0);

  /* frequency adjustment, AM (150 kHz - 21850 kHz), default startFreqAM */
  /* IC supports: B20 = 150 kHz - 21.85 MHz; C40+ = 150 kHz - 26.1 MHz   */
  adj_am_tune = gtk_adjustment_new( (float)(startFreqAM), 153.0, 26100.0, 1.0, 3.0, 0.0);

  /* horizontal scrollbars for adjustments */
  hscrollbar_vol = gtk_hscale_new(GTK_ADJUSTMENT(adj_vol));

  /* -------------------------------
      GTk signales 
     ------------------------------- */

  /* FM radio buttons */
  gtk_signal_connect_object(GTK_OBJECT(button_fm_start), "clicked",
                           GTK_SIGNAL_FUNC(fm_start_cb), GTK_OBJECT(window));

  gtk_signal_connect(GTK_OBJECT(button_fm_seek_down), "clicked",
                           GTK_SIGNAL_FUNC(fm_seek_cb), "seek down");
  gtk_signal_connect(GTK_OBJECT(button_fm_seek_up), "clicked",
                           GTK_SIGNAL_FUNC(fm_seek_cb), "seek up");

  gtk_signal_connect_object(GTK_OBJECT(button_fm_autoscan), "clicked",
                           GTK_SIGNAL_FUNC(fm_autoscan_cb), GTK_OBJECT(window));

  gtk_signal_connect_object(GTK_OBJECT(button_rds), "clicked",
                           GTK_SIGNAL_FUNC(fm_rds_cb), GTK_OBJECT(window));

  gtk_signal_connect_object(GTK_OBJECT(button_fm_smeter_update), "clicked",
                           GTK_SIGNAL_FUNC(fm_smeter_update_cb), GTK_OBJECT(window));

  //gtk_signal_connect_object(GTK_OBJECT(button_fm_band_jp), "clicked",
  //                         GTK_SIGNAL_FUNC(show_about_dialog), GTK_OBJECT(window));

  //gtk_signal_connect_object(GTK_OBJECT(button_fm_band_oirt), "clicked",
  //                         GTK_SIGNAL_FUNC(show_preferences), GTK_OBJECT(window));

  /* AM radio buttons */
  gtk_signal_connect_object(GTK_OBJECT(button_am_start), "clicked",
                           GTK_SIGNAL_FUNC(am_start_cb), GTK_OBJECT(window));

  gtk_signal_connect(GTK_OBJECT(button_am_seek_down), "clicked",
                           GTK_SIGNAL_FUNC(am_seek_cb), "seek down");

  gtk_signal_connect(GTK_OBJECT(button_am_seek_up), "clicked",
                           GTK_SIGNAL_FUNC(am_seek_cb), "seek up");

  gtk_signal_connect(GTK_OBJECT(button_am_tune_1up), "clicked",
                           GTK_SIGNAL_FUNC(am_tune_khz_cb), GINT_TO_POINTER(+1));

  gtk_signal_connect(GTK_OBJECT(button_am_tune_1down), "clicked",
                           GTK_SIGNAL_FUNC(am_tune_khz_cb), GINT_TO_POINTER(-1));

  gtk_signal_connect(GTK_OBJECT(button_am_tune_5up), "clicked",
                           GTK_SIGNAL_FUNC(am_tune_khz_cb), GINT_TO_POINTER(+5));

  gtk_signal_connect(GTK_OBJECT(button_am_tune_5down), "clicked",
                           GTK_SIGNAL_FUNC(am_tune_khz_cb), GINT_TO_POINTER(-5));

  gtk_signal_connect(GTK_OBJECT(button_am_tune_9up), "clicked",
                           GTK_SIGNAL_FUNC(am_tune_khz_cb), GINT_TO_POINTER(+9));

  gtk_signal_connect(GTK_OBJECT(button_am_tune_9down), "clicked",
                           GTK_SIGNAL_FUNC(am_tune_khz_cb), GINT_TO_POINTER(-9));

  gtk_signal_connect_object(GTK_OBJECT(button_am_autoscan), "clicked",
                           GTK_SIGNAL_FUNC(am_autoscan_cb), GTK_OBJECT(window));

  gtk_signal_connect_object(GTK_OBJECT(button_am_smeter_update), "clicked",
                           GTK_SIGNAL_FUNC(am_smeter_update_cb), GTK_OBJECT(window));

  /* general buttons */
  gtk_signal_connect_object(GTK_OBJECT(button_mute), "clicked",
                           GTK_SIGNAL_FUNC(vol_mute_cb), GTK_OBJECT(window));

  gtk_signal_connect_object(GTK_OBJECT(button_exit), "clicked",
                           GTK_SIGNAL_FUNC(exit_cb), GTK_OBJECT(window));
  gtk_signal_connect_object(GTK_OBJECT(button_quit), "clicked",
                           GTK_SIGNAL_FUNC(ende_cb), GTK_OBJECT(window));

  gtk_signal_connect(GTK_OBJECT(window), "destroy",
                     GTK_SIGNAL_FUNC(ende_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(adj_vol), "value_changed",
                     GTK_SIGNAL_FUNC(scrollbar_cb_vol), "Scrollbar Volume");

  /* -------------------------------
      status bar 
     ------------------------------- */
  statusbar = gtk_statusbar_new();

  /* set GTK window parameters  */
  gtk_window_set_title(GTK_WINDOW(window), "Si4735 FM/AM Radio");	// windows title
  gtk_window_set_default_size(GTK_WINDOW(window), 380, 150);		// default size
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);	// position
  gtk_container_set_border_width(GTK_CONTAINER(window), 3);		// border
  /* set app icon */
  gtk_window_set_icon(GTK_WINDOW(window), gdk_pixbuf_new_from_xpm_data((const char **)bmp_si4735));

  /* ************************* */
  /* *** pack GUI elements *** */
  /* ************************* */

  /* -------------------------------
      FM notebook page / tab
     ------------------------------- */

  /* RDS info */
  vbox = gtk_vbox_new(FALSE, 5);

  /* PI */
  hbox = gtk_hbox_new(FALSE, 5);
  label_rds_pi = gtk_label_new ("PI: ");
  gtk_label_set_width_chars(GTK_LABEL(label_rds_pi), 4);
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_pi, FALSE, FALSE, 4);
  entry_rds_pi = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_pi), "* PI *");
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_pi),8);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_pi), 7);
  gtk_entry_set_has_frame(GTK_ENTRY(entry_rds_pi),TRUE);
  g_object_set(entry_rds_pi, "editable", FALSE, NULL);
  //gtk_widget_set_sensitive(entry_rds_pi, _isSensitive);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_pi, FALSE, FALSE, 4);
  /* PS */
  label_rds_psn = gtk_label_new ("PS: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_psn, FALSE, FALSE, 4);
  entry_rds_psn = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_psn), "*RDS PS*");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_psn), 9);
  g_object_set(entry_rds_psn, "editable", FALSE, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_psn, FALSE, FALSE, 4);
  /* AF */
  label_rds_af = gtk_label_new ("AF: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_af, FALSE, FALSE, 4);
  entry_rds_af = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_af), "*** AF ***");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_af), 14);
  g_object_set(entry_rds_af, "editable", FALSE, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_af, TRUE, TRUE, 4);
  /* Time */
  label_rds_time = gtk_label_new ("UTC/Loc: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_time, FALSE, FALSE, 4);
  entry_rds_time = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_time), "*** Zeit ***");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_time), 14);
  g_object_set(entry_rds_time, "editable", FALSE, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_time, TRUE, TRUE, 4);
  /* RDS start/stop button */
  gtk_box_pack_start(GTK_BOX(hbox), button_rds, TRUE, TRUE, 5);
  /* put hbox in vbox */
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);

  /* RadioText part A + PTY */
  hbox = gtk_hbox_new(FALSE, 5);
  label_rds_rt_a = gtk_label_new ("RT A: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_rt_a, FALSE, FALSE, 4);
  entry_rds_rt_a = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_rt_a), "*** RDS Radio Text ***");
  g_object_set(entry_rds_rt_a, "editable", FALSE, NULL);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_rt_a), 63);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_rt_a, FALSE, FALSE, 4);
  label_rds_pty = gtk_label_new ("PTY: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_pty, FALSE, FALSE, 4);
  entry_rds_pty = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_pty), "*** PTY ***");
  g_object_set(entry_rds_pty, "editable", FALSE, NULL);
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_pty),11);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_pty),11);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_pty, FALSE, FALSE, 4);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);

  /* RadioText part B + MS + TP + TA */
  hbox = gtk_hbox_new(FALSE, 5);
  label_rds_rt_b = gtk_label_new ("RT B: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_rt_b, FALSE, FALSE, 4);
  entry_rds_rt_b = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_rt_b), "*** RDS Radio Text ***");
  g_object_set(entry_rds_rt_b, "editable", FALSE, NULL);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_rt_b), 63);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_rt_b, FALSE, FALSE, 4);

  /* TA */
  entry_rds_ta = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_ta), "TA");
  g_object_set(entry_rds_ta, "editable", FALSE, NULL);
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_ta),2);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_ta),2);
  gtk_box_pack_end(GTK_BOX(hbox), entry_rds_ta, FALSE, FALSE, 4);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);
  /* TP */
  entry_rds_tp = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_tp), "TP");
  g_object_set(entry_rds_tp, "editable", FALSE, NULL);
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_tp),2);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_tp),2);
  gtk_box_pack_end(GTK_BOX(hbox), entry_rds_tp, FALSE, FALSE, 4);
  /* MS */
  entry_rds_ms = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_ms), "M/S");
  g_object_set(entry_rds_ms, "editable", FALSE, NULL);
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_ms),3);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_ms),3);
  gtk_box_pack_end(GTK_BOX(hbox), entry_rds_ms, FALSE, FALSE, 4);

  /* pack RDS boxes */
  gtk_container_border_width (GTK_CONTAINER(frame_fm_rds), 5);
  gtk_container_add(GTK_CONTAINER(frame_fm_rds), vbox);
  gtk_box_pack_start(GTK_BOX(main_vbox_fm), frame_fm_rds, FALSE, FALSE, 0);

  /* tuning */
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);
  /* frequency */
  label = gtk_label_new ("Freq: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 4);

  entry_fm_freq = gtk_spin_button_new (GTK_ADJUSTMENT(adj_fm_tune), 0.01, 2);

  gtk_tooltips_set_tip (GTK_TOOLTIPS (entry_tips), entry_fm_freq,
                           _("Frequenz in MHz, Eingabe direkt, durch Drücken der Pfeile oder mit den Cursortasten bzw. PageUp/PageDown"), "");

  gtk_signal_connect (GTK_OBJECT (adj_fm_tune), "value_changed",
                      GTK_SIGNAL_FUNC (change_fm_digits),
                      (gpointer) entry_fm_freq);

  gtk_box_pack_start(GTK_BOX(hbox), entry_fm_freq, FALSE, FALSE, 5);

  label = gtk_label_new ("MHz");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);

  /* BTN: seek buttons */
  gtk_box_pack_start(GTK_BOX(hbox), button_fm_seek_down, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(hbox), button_fm_seek_up, FALSE, FALSE, 2);
  /* BTN: autoscan */
  gtk_box_pack_start(GTK_BOX(hbox), button_fm_autoscan, FALSE, FALSE, 2);
  /* BTN: FM band selection */
  //gtk_box_pack_end(GTK_BOX(hbox), button_fm_band_oirt, FALSE, FALSE, 2);
  //gtk_box_pack_end(GTK_BOX(hbox), button_fm_band_jp, FALSE, FALSE, 2);
  //gtk_box_pack_end(GTK_BOX(hbox), button_fm_band_norm, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);

  /* station preset buttons */
  hbox = gtk_hbox_new(FALSE, 5);

  button_fm_preset_bank  = gtk_button_new_with_label( "MEM 1" );

  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_preset_bank,
                           "Schaltet zwischen den Speicherbänken mit je 6 Plätzen durch", "");
  
  gtk_signal_connect_object(GTK_OBJECT(button_fm_preset_bank), "clicked",
                            GTK_SIGNAL_FUNC(fm_preset_change_bank_cb), GTK_OBJECT(window));

  gtk_box_pack_start(GTK_BOX(hbox), button_fm_preset_bank, FALSE, FALSE, 2);

  int p;
  for ( p = 0; p < 6; p++ )
  {
    char pLabel[12];
    char tLabel[32];

    /* get frequency */
    if ( FM_Preset[p].freq > 0 )
    {
      if ( strlen(FM_Preset[p].station_name) > 2 )
      {
        sprintf(pLabel, "%s", FM_Preset[p].station_name );
      }
      else
      {
        sprintf(pLabel, "%.2f MHz", (double)FM_Preset[p].freq/100 );
      }
      
      sprintf(tLabel, "P %d - %.2f MHz - %s", (p+1), (double)FM_Preset[p].freq/100, FM_Preset[p].station_name );
    }
    else
    {
      sprintf(pLabel, "- %d -", (p+1));
      sprintf(tLabel, "FM Speicherplatz #%d -", (p+1));
    }

    button_fm_preset[p]  = gtk_button_new_with_label( pLabel );

    gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_preset[p],
                             tLabel, "");

    gtk_signal_connect(GTK_OBJECT(button_fm_preset[p]), "button-press-event", 
			   GTK_SIGNAL_FUNC(fm_preset_onButtonPressed), (void *)p);

    gtk_box_pack_start(GTK_BOX(hbox), button_fm_preset[p], TRUE, TRUE, 2);
  } 
  
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);

  gtk_container_add(GTK_CONTAINER(frame_fm_tune), vbox);
  gtk_container_border_width (GTK_CONTAINER(frame_fm_tune), 5);
  gtk_box_pack_start(GTK_BOX(main_vbox_fm), frame_fm_tune, FALSE, FALSE, 0);

  /* S-meter info */
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);

  /* FM init button */
  gtk_box_pack_start(GTK_BOX(hbox), button_fm_start, FALSE, FALSE, 0);

  /* smeter update button */
  gtk_box_pack_start(GTK_BOX(hbox), button_fm_smeter_update, FALSE, FALSE, 5);

  /* RSSI */
  label_fm_smeter_rssi = gtk_label_new ("RSSI: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_fm_smeter_rssi, FALSE, FALSE, 4);
  entry_fm_smeter_rssi = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_fm_smeter_rssi), "*** RSSI ***");
  gtk_entry_set_max_length (GTK_ENTRY(entry_fm_smeter_rssi),12);
  gtk_entry_set_has_frame(GTK_ENTRY(entry_fm_smeter_rssi),TRUE);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_fm_smeter_rssi), 10);
  g_object_set(entry_fm_smeter_rssi, "editable", FALSE, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), entry_fm_smeter_rssi, TRUE, TRUE, 4);
  /* SNR */
  label_fm_smeter_snr = gtk_label_new ("SNR: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_fm_smeter_snr, FALSE, FALSE, 4);
  entry_fm_smeter_snr = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_fm_smeter_snr), "*** SNR ***");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_fm_smeter_snr), 10);
  gtk_box_pack_start(GTK_BOX(hbox), entry_fm_smeter_snr, TRUE, TRUE, 4);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);
  /* pack S-meter boxes */
  gtk_container_border_width (GTK_CONTAINER(frame_fm_smeter), 5);
  gtk_container_add(GTK_CONTAINER(frame_fm_smeter), vbox);

  gtk_box_pack_start(GTK_BOX(main_vbox_fm), frame_fm_smeter, FALSE, FALSE, 0);

  /* -------------------------------
      AM notebook page / tab
     ------------------------------- */

  /* tuning */
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);

  /* frequency */
  label = gtk_label_new ("Freq: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 4);

  entry_am_freq = gtk_spin_button_new(GTK_ADJUSTMENT(adj_am_tune), 1.0, 0);

  gtk_tooltips_set_tip (GTK_TOOLTIPS (entry_tips), entry_am_freq,
                           "Frequenz in kHz, Eingabe direkt, durch Drücken der Pfeile oder mit den Cursortasten bzw. PageUp/PageDown", "");

  gtk_signal_connect (GTK_OBJECT (adj_am_tune), "value_changed",
                      GTK_SIGNAL_FUNC (change_am_digits),
                      (gpointer) entry_am_freq);

  gtk_box_pack_start(GTK_BOX(hbox), entry_am_freq, FALSE, FALSE, 5);

  label = gtk_label_new ("kHz");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);


  /* +/- 1, 5, 9, 10 kHz buttons */
  gtk_box_pack_start(GTK_BOX(hbox), button_am_tune_1down, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(hbox), button_am_tune_1up, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(hbox), button_am_tune_5down, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(hbox), button_am_tune_5up, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(hbox), button_am_tune_9down, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(hbox), button_am_tune_9up, FALSE, FALSE, 2);

  /* seek buttons */
  gtk_box_pack_end(GTK_BOX(hbox), button_am_autoscan, FALSE, FALSE, 2);
  gtk_box_pack_end(GTK_BOX(hbox), button_am_seek_up, FALSE, FALSE, 2);
  gtk_box_pack_end(GTK_BOX(hbox), button_am_seek_down, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);

  gtk_container_add(GTK_CONTAINER(frame_am_tune), vbox);
  gtk_container_border_width (GTK_CONTAINER(frame_am_tune), 5);
  gtk_box_pack_start(GTK_BOX(main_vbox_am), frame_am_tune, FALSE, FALSE, 0);

  /* station preset buttons */
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);

  button_am_preset_bank  = gtk_button_new_with_label( "MEM 1" );

  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_preset_bank,
                           "Schaltet zwischen den Speicherbänken mit je 6 Plätze durch", "");
  
  gtk_signal_connect_object(GTK_OBJECT(button_am_preset_bank), "clicked",
                            GTK_SIGNAL_FUNC(am_preset_change_bank_cb), GTK_OBJECT(window));

  gtk_box_pack_start(GTK_BOX(hbox), button_am_preset_bank, FALSE, FALSE, 2);

  for ( p = 0; p < 6; p++ )
  {
    char pLabel[12];
    char tLabel[32];

    /* get frequency */
    if ( AM_Preset[p].freq > 0 )
    {
      if ( strlen(AM_Preset[p].station_name) > 2 )
      {
        sprintf(pLabel, "%s", AM_Preset[p].station_name );
      }
      else
      {
        sprintf(pLabel, "%d kHz", AM_Preset[p].freq );
      }
      
      sprintf(tLabel, "P %d - %d kHz - %s", (p+1), AM_Preset[p].freq, AM_Preset[p].station_name );
    }
    else
    {
      sprintf(pLabel, "- %d -", (p+1));
      sprintf(tLabel, "AM Speicherplatz #%d -", (p+1));
    }

    button_am_preset[p]  = gtk_button_new_with_label( pLabel );

    gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_preset[p],
                             tLabel, "");

    gtk_signal_connect(GTK_OBJECT(button_am_preset[p]), "button-press-event", 
			   GTK_SIGNAL_FUNC(am_preset_onButtonPressed), (void *)p);

    gtk_box_pack_start(GTK_BOX(hbox), button_am_preset[p], TRUE, TRUE, 2);
  } 
  
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);

  gtk_container_add(GTK_CONTAINER(frame_am_preset), vbox);
  gtk_container_border_width (GTK_CONTAINER(frame_am_preset), 5);
  gtk_box_pack_start(GTK_BOX(main_vbox_am), frame_am_preset, FALSE, FALSE, 0);

  /* freq band selection */
  vbox = gtk_vbox_new(FALSE, 5);

  int arr_len = sizeof(AM_bands)/sizeof(AM_bands[0]);

  for ( p = 0; p < arr_len; p++ )
  {
    if ( ((float)p/8 - rint((float)p/8)) == 0) 
    {
      bbox = gtk_hbutton_box_new();
      gtk_container_set_border_width (GTK_CONTAINER (bbox), 2);
      gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_START);
      gtk_box_set_spacing (GTK_BOX (bbox), 5);
      // gtk_button_box_set_child_size (GTK_BUTTON_BOX (bbox), 70, 10);
    } 

    char pLabel[8];
    char tLabel[32];

    if ( AM_bands[p].wavelength == 1500 )
    {
      sprintf(pLabel, "%s", "LW");
      sprintf(tLabel, "%s (%d - %d kHz)", "LW", AM_bands[p].band_bottom, AM_bands[p].band_top);
    }
    else if ( AM_bands[p].wavelength == 300 )
    {
      sprintf(pLabel, "%s", "MW");
      sprintf(tLabel, "%s (%d - %d kHz)", "MW", AM_bands[p].band_bottom, AM_bands[p].band_top);
    }
    else
    {
      sprintf(pLabel, "%d m", AM_bands[p].wavelength);
      sprintf(tLabel, "%d m (%d - %d kHz)", AM_bands[p].wavelength, AM_bands[p].band_bottom, AM_bands[p].band_top);
    }
      
    button_am_band[p]  = gtk_button_new_with_label( pLabel );
    // gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_am_band[p]), FALSE);
    
    gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_am_band[p],
                             tLabel, "");

    gtk_signal_connect(GTK_OBJECT(button_am_band[p]), "clicked",
                           GTK_SIGNAL_FUNC(am_freq_band_cb), (void *)p);

    gtk_box_pack_start(GTK_BOX(bbox), button_am_band[p], TRUE, TRUE, 2);

    if ( ((p > 0) && ( (float)p/7 - rint((float)p/7) == 0 )) || ((p + 1) == arr_len) )
    {
      gtk_box_pack_start(GTK_BOX(vbox), bbox, FALSE, FALSE, 5);
    }
  } 

  gtk_container_add(GTK_CONTAINER(frame_am_band), vbox);
  gtk_container_border_width (GTK_CONTAINER(frame_am_band), 5);
  gtk_box_pack_start(GTK_BOX(main_vbox_am), frame_am_band, FALSE, FALSE, 0);

  /* S-meter info */
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);

  gtk_box_pack_start(GTK_BOX(hbox), button_am_start, TRUE, TRUE, 5);
  
  /* smeter update button */
  gtk_box_pack_start(GTK_BOX(hbox), button_am_smeter_update, FALSE, FALSE, 5);

  /* RSSI */
  label_am_smeter_rssi = gtk_label_new ("RSSI: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_am_smeter_rssi, FALSE, FALSE, 4);
  entry_am_smeter_rssi = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_am_smeter_rssi), "*** RSSI ***");
  gtk_entry_set_max_length (GTK_ENTRY(entry_am_smeter_rssi),12);
  gtk_entry_set_has_frame(GTK_ENTRY(entry_am_smeter_rssi),TRUE);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_am_smeter_rssi), 6);
  g_object_set(entry_am_smeter_rssi, "editable", FALSE, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), entry_am_smeter_rssi, TRUE, TRUE, 4);

  /* SNR */
  label_am_smeter_snr = gtk_label_new ("SNR: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_am_smeter_snr, FALSE, FALSE, 4);
  entry_am_smeter_snr = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_am_smeter_snr), "*** SNR ***");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_am_smeter_snr), 6);
  gtk_box_pack_start(GTK_BOX(hbox), entry_am_smeter_snr, TRUE, TRUE, 4);

  /* CAP */
  label_am_smeter_cap = gtk_label_new ("Kap: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_am_smeter_cap, FALSE, FALSE, 4);
  entry_am_smeter_cap = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_am_smeter_cap), "*** CAP ***");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_am_smeter_cap), 6);
  gtk_box_pack_start(GTK_BOX(hbox), entry_am_smeter_cap, TRUE, TRUE, 4);

  /* IND */
  label_am_smeter_ind = gtk_label_new ("Ind: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_am_smeter_ind, FALSE, FALSE, 4);
  entry_am_smeter_ind = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_am_smeter_ind), "*** IND ***");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_am_smeter_ind), 7);
  gtk_box_pack_start(GTK_BOX(hbox), entry_am_smeter_ind, TRUE, TRUE, 4);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);

  /* pack S-meter boxes */
  gtk_container_border_width (GTK_CONTAINER(frame_am_smeter), 5);
  gtk_container_add(GTK_CONTAINER(frame_am_smeter), vbox);
  gtk_box_pack_start(GTK_BOX(main_vbox_am), frame_am_smeter, FALSE, FALSE, 0);

  /* volume */ 
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), hscrollbar_vol, TRUE, TRUE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_mute, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_quit, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_exit, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame_vol), vbox);
  gtk_container_border_width (GTK_CONTAINER(frame_vol), 5);
  gtk_box_pack_start(GTK_BOX(global_vbox), frame_vol, FALSE, FALSE, 0);

  /* -------------------------------
      GPIO notebook page / tab
     ------------------------------- */

  /* ### GUI frames - GPIO part ### */

  /* create frame for GPIO 1 */
  frame_gpio1 = gtk_frame_new("GPIO 1");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_gpio1), GTK_SHADOW_ETCHED_IN);

  /* create frame for GPIO 2 */
  frame_gpio2 = gtk_frame_new("GPIO 2");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_gpio2), GTK_SHADOW_ETCHED_IN);

  /* create frame for GPIO 3 */
  frame_gpio3 = gtk_frame_new("GPIO 3");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_gpio3), GTK_SHADOW_ETCHED_IN);

  /* ### main vbox  and hbox ### */
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);

  /* GPIO1 low button */
  button_gpio1_1  = gtk_button_new_with_label("GPIO #1 - low");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_gpio1_1,
                           "GPIO1", "");

  gtk_signal_connect(GTK_OBJECT(button_gpio1_1), "clicked",
                           GTK_SIGNAL_FUNC(gpio_cb), GINT_TO_POINTER(1));

  /* GPIO1 high button */
  button_gpio1_2  = gtk_button_new_with_label("GPIO #1 - high");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_gpio1_2,
                           "GPIO2", "");

  gtk_signal_connect(GTK_OBJECT(button_gpio1_2), "clicked",
                           GTK_SIGNAL_FUNC(gpio_cb), GINT_TO_POINTER(0));


  /* GPIO1 100ms puls x 1 */
  button_gpio1_1p  = gtk_button_new_with_label("GPIO #1 - 1 pulse");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_gpio1_1p,
                           "GPIO1 1p", "");

  gtk_signal_connect(GTK_OBJECT(button_gpio1_1p), "clicked",
                           GTK_SIGNAL_FUNC(gpio_pulse_cb), GINT_TO_POINTER(1));

  /* GPIO1 100ms puls x 2 */
  button_gpio1_2p  = gtk_button_new_with_label("GPIO #1 - 2 pulses");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_gpio1_2p,
                           "GPIO1 2p", "");

  gtk_signal_connect(GTK_OBJECT(button_gpio1_2p), "clicked",
                           GTK_SIGNAL_FUNC(gpio_pulse_cb), GINT_TO_POINTER(2));

  /* GPIO1 100ms puls x 3 */
  button_gpio1_3p  = gtk_button_new_with_label("GPIO #1 - 3 pulses");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_gpio1_3p,
                           "GPIO1 3p", "");

  gtk_signal_connect(GTK_OBJECT(button_gpio1_3p), "clicked",
                           GTK_SIGNAL_FUNC(gpio_pulse_cb), GINT_TO_POINTER(3));

  /* GPIO1 100ms puls x 4 */
  button_gpio1_4p  = gtk_button_new_with_label("GPIO #1 - 4 pulses");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_gpio1_4p,
                           "GPIO1 4p", "");

  gtk_signal_connect(GTK_OBJECT(button_gpio1_4p), "clicked",
                           GTK_SIGNAL_FUNC(gpio_pulse_cb), GINT_TO_POINTER(4));

  /* GPIO 1 */
  label = gtk_label_new ("GPIO #1: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 4);

  gtk_box_pack_start(GTK_BOX(hbox), button_gpio1_1, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_gpio1_2, FALSE, FALSE, 5);

  gtk_box_pack_start(GTK_BOX(hbox), button_gpio1_1p, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_gpio1_2p, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_gpio1_3p, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_gpio1_4p, FALSE, FALSE, 5);

  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);

  gtk_container_add(GTK_CONTAINER(frame_gpio1), vbox);
  gtk_container_border_width (GTK_CONTAINER(frame_gpio1), 5);
  gtk_box_pack_start(GTK_BOX(main_vbox_gp), frame_gpio1, FALSE, FALSE, 0);

//  gtk_box_pack_start(GTK_BOX(global_vbox), frame_gp, FALSE, FALSE, 0);

  /* statusbar? */
  gtk_box_pack_start(GTK_BOX(global_vbox), statusbar, FALSE, TRUE, 1);
  
  /* show Gtk window etc... */
  gtk_widget_show_all(window);

  /* check if radio is already running */
  FM_Tune_Status(3);  
  Delay(110);
  AM_Tune_Status(3);
  Delay(110);
  
  if ( isModeFM )
  { 
    char *gstr = _("FM Mode running ....\0");
    gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), gstr), gstr);
    FM_SMeter();
  }
  else if ( isModeAM )
  { 
    char *gstr = _("AM Mode running ....\0");
    gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), gstr), gstr);
    gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), 1);
    AM_SMeter();
  }
  else
  {
    char *gstr = _("Radio in Power Down state ....\0");
    gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), gstr), gstr);
  }
  
  /* start Gtk main loop */
  gtk_main();

  return EXIT_SUCCESS;
}
